# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections.abc import AsyncIterable
from typing import AsyncGenerator
import time

from fastapi import APIRouter, FastAPI, Request, Response
from fastapi.exceptions import RequestValidationError
from pydantic import TypeAdapter, ValidationError
from starlette.responses import StreamingResponse

from kserve.protocol.rest.openai.types.openapi import (
    CreateChatCompletionRequest,
    CreateCompletionRequest,
    CreateEmbeddingRequest,
    ListModelsResponse,
    Model,
)

from ....errors import ModelNotReady
from .dataplane import OpenAIDataPlane
from .errors import OpenAIError, openai_error_handler

OPENAI_ROUTE_PREFIX = os.environ.get("KSERVE_OPENAI_ROUTE_PREFIX", "/openai")

if len(OPENAI_ROUTE_PREFIX) > 0 and not OPENAI_ROUTE_PREFIX.startswith("/"):
    OPENAI_ROUTE_PREFIX = f"/{OPENAI_ROUTE_PREFIX}"


CreateCompletionRequestAdapter = TypeAdapter(CreateCompletionRequest)
ChatCompletionRequestAdapter = TypeAdapter(CreateChatCompletionRequest)
EmbeddingRequestAdapter = TypeAdapter(CreateEmbeddingRequest)


class OpenAIEndpoints:
    def __init__(self, dataplane: OpenAIDataPlane):
        self.dataplane = dataplane
        self.start_time = int(time.time())

    async def create_completion(
        self,
        raw_request: Request,
        request_body: CreateCompletionRequest,
        response: Response,
    ) -> Response:
        """Create completion handler.

        Args:
            raw_request (Request): fastapi request object,
            request_body (CompletionCreateParams): Completion params body.
            response (Response): fastapi response object

        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = CreateCompletionRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError(errors=e.errors())
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        completion = await self.dataplane.create_completion(
            model_name=model_name,
            request=params,
            headers=raw_request.headers,
            response=response,
        )
        if isinstance(completion, AsyncIterable):

            async def stream_results() -> AsyncGenerator[str, None]:
                async for partial_completion in completion:
                    yield f"data: {partial_completion.model_dump_json()}\n\n"
                yield "data: [DONE]\n\n"

            return StreamingResponse(stream_results(), media_type="text/event-stream")
        else:
            return completion

    async def create_chat_completion(
        self,
        raw_request: Request,
        request_body: CreateChatCompletionRequest,
        response: Response,
    ) -> Response:
        """Create chat completion handler.

        Args:
            raw_request (Request): fastapi request object,
            request_body (ChatCompletionRequestAdapter): Chat completion params body.
            response (Response): fastapi response object

        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = ChatCompletionRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError(errors=e.errors())
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        request_headers = raw_request.headers
        completion = await self.dataplane.create_chat_completion(
            model_name=model_name,
            request=request_body,
            headers=request_headers,
            response=response,
        )
        if isinstance(completion, AsyncIterable):

            async def stream_results() -> AsyncGenerator[str, None]:
                async for chunk in completion:
                    yield f"data: {chunk.model_dump_json()}\n\n"
                yield "data: [DONE]\n\n"

            return StreamingResponse(stream_results(), media_type="text/event-stream")
        else:
            return completion

    async def create_embedding(
        self,
        raw_request: Request,
        request_body: CreateEmbeddingRequest,
        response: Response,
    ) -> Response:
        """Create embedding handler.

        Args:
            raw_request (Request): fastapi request object,
            request_body (CreateEmbeddingRequest): Embedding params body.
            response (Response): fastapi response object

        Returns:
            InferenceResponse: Inference response object.
        """
        try:
            params = EmbeddingRequestAdapter.validate_python(request_body)
        except ValidationError as e:
            raise RequestValidationError(errors=e.errors())
        params = request_body
        model_name = params.model
        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        request_headers = raw_request.headers
        return await self.dataplane.create_embedding(
            model_name=model_name,
            request=request_body,
            headers=request_headers,
            response=response,
        )

    async def models(
        self,
    ) -> ListModelsResponse:
        """Create chat completion handler.

        Args:
            raw_request (Request): fastapi request object,

        Returns:
            ListModelsResponse: Model response object.
        """
        models = await self.dataplane.models()
        return ListModelsResponse(
            object="list",
            data=[
                Model(
                    object="model", id=model.name, created=self.start_time, owned_by=""
                )
                for model in models
            ],
        )

    async def health(self, model_name: str):
        await self.dataplane.model_ready(model_name)


def register_openai_endpoints(app: FastAPI, dataplane: OpenAIDataPlane):
    endpoints = OpenAIEndpoints(dataplane)
    openai_router = APIRouter(prefix=OPENAI_ROUTE_PREFIX, tags=["OpenAI"])
    openai_router.add_api_route(
        r"/v1/completions",
        endpoints.create_completion,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/chat/completions",
        endpoints.create_chat_completion,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/embeddings",
        endpoints.create_embedding,
        methods=["POST"],
        response_model_exclude_none=True,
        response_model_exclude_unset=True,
    )
    openai_router.add_api_route(
        r"/v1/models",
        endpoints.models,
        methods=["GET"],
    )
    openai_router.add_api_route(
        r"/v1/models/{model_name}", endpoints.health, methods=["GET"]
    )
    app.include_router(openai_router)
    app.add_exception_handler(OpenAIError, openai_error_handler)



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1alpha1InferenceStep(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'condition': 'str',
        'data': 'str',
        'dependency': 'str',
        'name': 'str',
        'node_name': 'str',
        'service_name': 'str',
        'service_url': 'str',
        'weight': 'int'
    }

    attribute_map = {
        'condition': 'condition',
        'data': 'data',
        'dependency': 'dependency',
        'name': 'name',
        'node_name': 'nodeName',
        'service_name': 'serviceName',
        'service_url': 'serviceUrl',
        'weight': 'weight'
    }

    def __init__(self, condition=None, data=None, dependency=None, name=None, node_name=None, service_name=None, service_url=None, weight=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1InferenceStep - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._condition = None
        self._data = None
        self._dependency = None
        self._name = None
        self._node_name = None
        self._service_name = None
        self._service_url = None
        self._weight = None
        self.discriminator = None

        if condition is not None:
            self.condition = condition
        if data is not None:
            self.data = data
        if dependency is not None:
            self.dependency = dependency
        if name is not None:
            self.name = name
        if node_name is not None:
            self.node_name = node_name
        if service_name is not None:
            self.service_name = service_name
        if service_url is not None:
            self.service_url = service_url
        if weight is not None:
            self.weight = weight

    @property
    def condition(self):
        """Gets the condition of this V1alpha1InferenceStep.  # noqa: E501

        routing based on the condition  # noqa: E501

        :return: The condition of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._condition

    @condition.setter
    def condition(self, condition):
        """Sets the condition of this V1alpha1InferenceStep.

        routing based on the condition  # noqa: E501

        :param condition: The condition of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._condition = condition

    @property
    def data(self):
        """Gets the data of this V1alpha1InferenceStep.  # noqa: E501

        request data sent to the next route with input/output from the previous step $request $response.predictions  # noqa: E501

        :return: The data of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._data

    @data.setter
    def data(self, data):
        """Sets the data of this V1alpha1InferenceStep.

        request data sent to the next route with input/output from the previous step $request $response.predictions  # noqa: E501

        :param data: The data of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._data = data

    @property
    def dependency(self):
        """Gets the dependency of this V1alpha1InferenceStep.  # noqa: E501

        to decide whether a step is a hard or a soft dependency in the Inference Graph  # noqa: E501

        :return: The dependency of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._dependency

    @dependency.setter
    def dependency(self, dependency):
        """Sets the dependency of this V1alpha1InferenceStep.

        to decide whether a step is a hard or a soft dependency in the Inference Graph  # noqa: E501

        :param dependency: The dependency of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._dependency = dependency

    @property
    def name(self):
        """Gets the name of this V1alpha1InferenceStep.  # noqa: E501

        Unique name for the step within this node  # noqa: E501

        :return: The name of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this V1alpha1InferenceStep.

        Unique name for the step within this node  # noqa: E501

        :param name: The name of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def node_name(self):
        """Gets the node_name of this V1alpha1InferenceStep.  # noqa: E501

        The node name for routing as next step  # noqa: E501

        :return: The node_name of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._node_name

    @node_name.setter
    def node_name(self, node_name):
        """Sets the node_name of this V1alpha1InferenceStep.

        The node name for routing as next step  # noqa: E501

        :param node_name: The node_name of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._node_name = node_name

    @property
    def service_name(self):
        """Gets the service_name of this V1alpha1InferenceStep.  # noqa: E501

        named reference for InferenceService  # noqa: E501

        :return: The service_name of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._service_name

    @service_name.setter
    def service_name(self, service_name):
        """Sets the service_name of this V1alpha1InferenceStep.

        named reference for InferenceService  # noqa: E501

        :param service_name: The service_name of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._service_name = service_name

    @property
    def service_url(self):
        """Gets the service_url of this V1alpha1InferenceStep.  # noqa: E501

        InferenceService URL, mutually exclusive with ServiceName  # noqa: E501

        :return: The service_url of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: str
        """
        return self._service_url

    @service_url.setter
    def service_url(self, service_url):
        """Sets the service_url of this V1alpha1InferenceStep.

        InferenceService URL, mutually exclusive with ServiceName  # noqa: E501

        :param service_url: The service_url of this V1alpha1InferenceStep.  # noqa: E501
        :type: str
        """

        self._service_url = service_url

    @property
    def weight(self):
        """Gets the weight of this V1alpha1InferenceStep.  # noqa: E501

        the weight for split of the traffic, only used for Split Router when weight is specified all the routing targets should be sum to 100  # noqa: E501

        :return: The weight of this V1alpha1InferenceStep.  # noqa: E501
        :rtype: int
        """
        return self._weight

    @weight.setter
    def weight(self, weight):
        """Sets the weight of this V1alpha1InferenceStep.

        the weight for split of the traffic, only used for Split Router when weight is specified all the routing targets should be sum to 100  # noqa: E501

        :param weight: The weight of this V1alpha1InferenceStep.  # noqa: E501
        :type: int
        """

        self._weight = weight

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1InferenceStep):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1InferenceStep):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import List
from urllib.parse import urlparse

import requests
from kubernetes import client, config

from ..models.v1alpha1_local_model_cache import V1alpha1LocalModelCache
from ..models.v1alpha1_local_model_node_group import V1alpha1LocalModelNodeGroup
from .creds_utils import set_gcs_credentials, set_s3_credentials, set_azure_credentials
from .watch import isvc_watch
from ..constants import constants
from ..models import V1alpha1InferenceGraph
from ..utils import utils


class KServeClient(object):

    def __init__(
        self,
        config_file=None,
        config_dict=None,
        context=None,  # pylint: disable=too-many-arguments
        client_configuration=None,
        persist_config=True,
    ):
        """
        KServe client constructor
        :param config_file: kubeconfig file, defaults to ~/.kube/config
        :param config_dict: Takes the config file as a dict.
        :param context: kubernetes context
        :param client_configuration: kubernetes configuration object
        :param persist_config:
        """
        if config_file or config_dict or not utils.is_running_in_k8s():
            if config_dict:
                config.load_kube_config_from_dict(
                    config_dict=config_dict,
                    context=context,
                    client_configuration=None,
                    persist_config=persist_config,
                )
            else:
                config.load_kube_config(
                    config_file=config_file,
                    context=context,
                    client_configuration=client_configuration,
                    persist_config=persist_config,
                )
        else:
            config.load_incluster_config()
        self.core_api = client.CoreV1Api()
        self.app_api = client.AppsV1Api()
        self.api_instance = client.CustomObjectsApi()
        self.hpa_v2_api = client.AutoscalingV2Api()

    def set_credentials(
        self,
        storage_type,
        namespace=None,
        credentials_file=None,
        service_account=constants.DEFAULT_SA_NAME,
        **kwargs,
    ):
        """
        Setup credentials for KServe.

        :param storage_type: Valid value: GCS or S3 (required)
        :param namespace: inference service deployment namespace
        :param credentials_file: the path for the credentials file.
        :param service_account: the name of service account.
        :param kwargs: Others parameters for each storage_type
        :return:
        """

        if namespace is None:
            namespace = utils.get_default_target_namespace()

        if storage_type.lower() == "gcs":
            if credentials_file is None:
                credentials_file = constants.GCS_DEFAULT_CREDS_FILE
            set_gcs_credentials(
                namespace=namespace,
                credentials_file=credentials_file,
                service_account=service_account,
            )
        elif storage_type.lower() == "s3":
            if credentials_file is None:
                credentials_file = constants.S3_DEFAULT_CREDS_FILE
            set_s3_credentials(
                namespace=namespace,
                credentials_file=credentials_file,
                service_account=service_account,
                **kwargs,
            )
        elif storage_type.lower() == "azure":
            if credentials_file is None:
                credentials_file = constants.AZ_DEFAULT_CREDS_FILE
            set_azure_credentials(
                namespace=namespace,
                credentials_file=credentials_file,
                service_account=service_account,
            )
        else:
            raise RuntimeError(
                "Invalid storage_type: %s, only support GCS, S3 and Azure\
                currently.\n"
                % storage_type
            )

    def create(
        self, inferenceservice, namespace=None, watch=False, timeout_seconds=600
    ):  # pylint:disable=inconsistent-return-statements
        """
        Create the inference service
        :param inferenceservice: inference service object
        :param namespace: defaults to current or default namespace
        :param watch: True to watch the created service until timeout elapsed or status is ready
        :param timeout_seconds: timeout seconds for watch, default to 600s
        :return: created inference service
        """

        version = inferenceservice.api_version.split("/")[1]

        if namespace is None:
            namespace = utils.get_isvc_namespace(inferenceservice)

        try:
            outputs = self.api_instance.create_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCESERVICE,
                inferenceservice,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->create_namespaced_custom_object:\
                 %s\n"
                % e
            )

        if watch:
            isvc_watch(
                name=outputs["metadata"]["name"],
                namespace=namespace,
                timeout_seconds=timeout_seconds,
            )
        else:
            return outputs

    def get(
        self,
        name=None,
        namespace=None,
        watch=False,
        timeout_seconds=600,
        version=constants.KSERVE_V1BETA1_VERSION,
    ):  # pylint:disable=inconsistent-return-statements
        """
        Get the inference service
        :param name: existing inference service name
        :param namespace: defaults to current or default namespace
        :param watch: True to watch the service until timeout elapsed or status is ready
        :param timeout_seconds: timeout seconds for watch, default to 600s
        :param version: api group version
        :return: inference service
        """

        if namespace is None:
            namespace = utils.get_default_target_namespace()

        if name:
            if watch:
                isvc_watch(
                    name=name, namespace=namespace, timeout_seconds=timeout_seconds
                )
            else:
                try:
                    return self.api_instance.get_namespaced_custom_object(
                        constants.KSERVE_GROUP,
                        version,
                        namespace,
                        constants.KSERVE_PLURAL_INFERENCESERVICE,
                        name,
                    )
                except client.rest.ApiException as e:
                    raise RuntimeError(
                        "Exception when calling CustomObjectsApi->get_namespaced_custom_object:\
                        %s\n"
                        % e
                    )
        else:
            if watch:
                isvc_watch(namespace=namespace, timeout_seconds=timeout_seconds)
            else:
                try:
                    return self.api_instance.list_namespaced_custom_object(
                        constants.KSERVE_GROUP,
                        version,
                        namespace,
                        constants.KSERVE_PLURAL_INFERENCESERVICE,
                    )
                except client.rest.ApiException as e:
                    raise RuntimeError(
                        "Exception when calling CustomObjectsApi->list_namespaced_custom_object:\
                        %s\n"
                        % e
                    )

    def patch(
        self, name, inferenceservice, namespace=None, watch=False, timeout_seconds=600
    ):  # pylint:disable=too-many-arguments,inconsistent-return-statements
        """
        Patch existing inference service
        :param name: existing inference service name
        :param inferenceservice: patched inference service
        :param namespace: defaults to current or default namespace
        :param watch: True to watch the patched service until timeout elapsed or status is ready
        :param timeout_seconds: timeout seconds for watch, default to 600s
        :return: patched inference service
        """

        version = inferenceservice.api_version.split("/")[1]
        if namespace is None:
            namespace = utils.get_isvc_namespace(inferenceservice)

        try:
            outputs = self.api_instance.patch_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCESERVICE,
                name,
                inferenceservice,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->patch_namespaced_custom_object:\
                 %s\n"
                % e
            )

        if watch:
            # Sleep 3 to avoid status still be True within a very short time.
            time.sleep(3)
            isvc_watch(
                name=outputs["metadata"]["name"],
                namespace=namespace,
                timeout_seconds=timeout_seconds,
            )
        else:
            return outputs

    def replace(
        self, name, inferenceservice, namespace=None, watch=False, timeout_seconds=600
    ):  # pylint:disable=too-many-arguments,inconsistent-return-statements
        """
        Replace the existing inference service
        :param name: existing inference service name
        :param inferenceservice: replacing inference service
        :param namespace: defaults to current or default namespace
        :param watch: True to watch the replaced service until timeout elapsed or status is ready
        :param timeout_seconds: timeout seconds for watch, default to 600s
        :return: replaced inference service
        """

        version = inferenceservice.api_version.split("/")[1]

        if namespace is None:
            namespace = utils.get_isvc_namespace(inferenceservice)

        if inferenceservice.metadata.resource_version is None:
            current_isvc = self.get(name, namespace=namespace)
            current_resource_version = current_isvc["metadata"]["resourceVersion"]
            inferenceservice.metadata.resource_version = current_resource_version

        try:
            outputs = self.api_instance.replace_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCESERVICE,
                name,
                inferenceservice,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->replace_namespaced_custom_object:\
                 %s\n"
                % e
            )

        if watch:
            isvc_watch(
                name=outputs["metadata"]["name"],
                namespace=namespace,
                timeout_seconds=timeout_seconds,
                generation=outputs["metadata"]["generation"],
            )
        else:
            return outputs

    def delete(self, name, namespace=None, version=constants.KSERVE_V1BETA1_VERSION):
        """
        Delete the inference service
        :param name: inference service name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        :return:
        """
        if namespace is None:
            namespace = utils.get_default_target_namespace()

        try:
            return self.api_instance.delete_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCESERVICE,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->delete_namespaced_custom_object:\
                 %s\n"
                % e
            )

    def is_isvc_ready(
        self, name, namespace=None, version=constants.KSERVE_V1BETA1_VERSION
    ):  # pylint:disable=inconsistent-return-statements
        """
        Check if the inference service is ready.
        :param version:
        :param name: inference service name
        :param namespace: defaults to current or default namespace
        :return:
        """
        kfsvc_status = self.get(name, namespace=namespace, version=version)
        if "status" not in kfsvc_status:
            return False
        status = "Unknown"
        for condition in kfsvc_status["status"].get("conditions", {}):
            if condition.get("type", "") == "Ready":
                status = condition.get("status", "Unknown")
                return status.lower() == "true"
        return False

    def wait_isvc_ready(
        self,
        name,
        namespace=None,  # pylint:disable=too-many-arguments
        watch=False,
        timeout_seconds=600,
        polling_interval=10,
        version=constants.KSERVE_V1BETA1_VERSION,
    ):
        """
        Waiting for inference service ready, print out the inference service if timeout.
        :param name: inference service name
        :param namespace: defaults to current or default namespace
        :param watch: True to watch the service until timeout elapsed or status is ready
        :param timeout_seconds: timeout seconds for waiting, default to 600s.
               Print out the InferenceService if timeout.
        :param polling_interval: The time interval to poll status
        :param version: api group version
        :return:
        """
        if watch:
            isvc_watch(name=name, namespace=namespace, timeout_seconds=timeout_seconds)
        else:
            for _ in range(round(timeout_seconds / polling_interval)):
                time.sleep(polling_interval)
                if self.is_isvc_ready(name, namespace=namespace, version=version):
                    return

            current_isvc = self.get(name, namespace=namespace, version=version)
            raise RuntimeError(
                "Timeout to start the InferenceService {}. \
                               The InferenceService is as following: {}".format(
                    name, current_isvc
                )
            )

    def create_trained_model(self, trainedmodel, namespace):
        """
        Create a trained model
        :param trainedmodel: trainedmodel object
        :param namespace: defaults to current or default namespace
        :return:
        """
        version = trainedmodel.api_version.split("/")[1]

        try:
            self.api_instance.create_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_TRAINEDMODEL,
                trainedmodel,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->create_namespaced_custom_object:\
                     %s\n"
                % e
            )

    def delete_trained_model(
        self, name, namespace=None, version=constants.KSERVE_V1ALPHA1_VERSION
    ):
        """
        Delete the trained model
        :param name: trained model name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        :return:
        """
        if namespace is None:
            namespace = utils.get_default_target_namespace()

        try:
            return self.api_instance.delete_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_TRAINEDMODEL,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->delete_namespaced_custom_object:\
                 %s\n"
                % e
            )

    def wait_model_ready(
        self,
        service_name,
        model_name,
        isvc_namespace=None,  # pylint:disable=too-many-arguments
        isvc_version=constants.KSERVE_V1BETA1_VERSION,
        cluster_ip=None,
        protocol_version="v1",
        timeout_seconds=600,
        polling_interval=10,
    ):
        """
        Waiting for model to be ready to service, print out trained model if timeout.
        :param service_name: inference service name
        :param model_name: trained model name
        :param isvc_namespace: defaults to current or default namespace of inference service
        :param isvc_version: api group version of inference service
        :param protocol_version: version of the dataplane protocol
        :param cluster_ip: ip of the kubernetes cluster
        :param timeout_seconds: timeout seconds for waiting, default to 600s.
          Print out the InferenceService if timeout.
        :param polling_interval: The time interval to poll status
        :return:
        """
        isvc = self.get(
            service_name,
            namespace=isvc_namespace,
            version=isvc_version,
        )

        host = urlparse(isvc["status"]["url"]).netloc
        headers = {"Host": host}

        for _ in range(round(timeout_seconds / polling_interval)):
            time.sleep(polling_interval)
            # Check model health API
            url = f"http://{cluster_ip}/{protocol_version}/models/{model_name}"
            if protocol_version.lower() == "v2":
                url = (
                    f"http://{cluster_ip}/{protocol_version}/models/{model_name}/ready"
                )
            response = requests.get(url, headers=headers).status_code
            if response == 200:
                return

        raise RuntimeError(
            f"InferenceService ({service_name}) has not loaded the \
                            model ({model_name}) before the timeout."
        )

    def create_inference_graph(
        self, inferencegraph: V1alpha1InferenceGraph, namespace: str = None
    ) -> object:
        """
        create a inference graph

        :param inferencegraph: inference graph object
        :param namespace: defaults to current or default namespace
        :return: created inference graph
        """
        version = inferencegraph.api_version.split("/")[1]
        if namespace is None:
            namespace = utils.get_ig_namespace(inferencegraph)

        try:
            outputs = self.api_instance.create_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCEGRAPH,
                inferencegraph,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->create_namespaced_custom_object:\
                 %s\n"
                % e
            )
        return outputs

    def delete_inference_graph(
        self,
        name: str,
        namespace: str = None,
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
    ):
        """
        Delete the inference graph

        :param name: inference graph name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        """
        if namespace is None:
            namespace = utils.get_default_target_namespace()

        try:
            self.api_instance.delete_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCEGRAPH,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->create_namespaced_custom_object:\
                 %s\n"
                % e
            )

    def get_inference_graph(
        self,
        name: str,
        namespace: str = None,
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
    ) -> object:
        """
        Get the inference graph

        :param name: existing inference graph name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        :return: inference graph
        """

        if namespace is None:
            namespace = utils.get_default_target_namespace()

        try:
            return self.api_instance.get_namespaced_custom_object(
                constants.KSERVE_GROUP,
                version,
                namespace,
                constants.KSERVE_PLURAL_INFERENCEGRAPH,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                "Exception when calling CustomObjectsApi->get_namespaced_custom_object:\
                 %s\n"
                % e
            )

    def is_ig_ready(
        self,
        name: str,
        namespace: str = None,
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
    ) -> bool:
        """
        Check if the inference graph is ready.

        :param name: inference graph name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        :return: true if inference graph is ready, else false.
        """
        if namespace is None:
            namespace = utils.get_default_target_namespace()

        ig: dict = self.get_inference_graph(name, namespace=namespace, version=version)
        for condition in ig.get("status", {}).get("conditions", {}):
            if condition.get("type", "") == "Ready":
                status = condition.get("status", "Unknown")
                return status.lower() == "true"
        return False

    def wait_ig_ready(
        self,
        name: str,
        namespace: str = None,
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
        timeout_seconds: int = 600,
        polling_interval: int = 10,
    ):
        """
        Wait for inference graph to be ready until timeout. Print out the inference graph if timeout.

        :param name: inference graph name
        :param namespace: defaults to current or default namespace
        :param version: api group version
        :param timeout_seconds: timeout seconds for waiting, default to 600s.
        :param polling_interval: The time interval to poll status
        :return:
        """
        for _ in range(round(timeout_seconds / polling_interval)):
            time.sleep(polling_interval)
            if self.is_ig_ready(name, namespace, version):
                return

        current_ig = self.get_inference_graph(
            name, namespace=namespace, version=version
        )
        raise RuntimeError(
            "Timeout to start the InferenceGraph {}. \
                            The InferenceGraph is as following: {}".format(
                name, current_ig
            )
        )

    def create_local_model_node_group(
        self, localmodelnodegroup: V1alpha1LocalModelNodeGroup
    ):
        """
        Create a local model node group

        :param localmodelnodegroup: local model node group object
        :return: created local model node group object
        """
        version = localmodelnodegroup.api_version.split("/")[1]

        try:
            output = self.api_instance.create_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELNODEGROUP,
                localmodelnodegroup,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->create_cluster_custom_object:{e}%s\n"
            ) from e
        return output

    def get_local_model_node_group(
        self, name: str, version: str = constants.KSERVE_V1ALPHA1_VERSION
    ) -> object:
        """
        Get the local model node group

        :param name: existing local model node group name
        :param version: api group version. Default to v1alpha
        :return: local model node group object
        """
        try:
            return self.api_instance.get_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELNODEGROUP,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->get_cluster_custom_object:{e}%s\n"
            ) from e

    def delete_local_model_node_group(
        self, name: str, version: str = constants.KSERVE_V1ALPHA1_VERSION
    ):
        """
        Delete the local model node group

        :param name: local model node group name
        :param version: api group version. Default to v1alpha
        """
        try:
            self.api_instance.delete_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELNODEGROUP,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->delete_cluster_custom_object:{e}%s\n"
            ) from e

    def create_local_model_cache(
        self, localmodelcache: V1alpha1LocalModelCache
    ) -> object:
        """
        Create a local model cache

        :param localmodelcache: local model cache object
        :return: created local model cache object
        """
        version = localmodelcache.api_version.split("/")[1]

        try:
            output = self.api_instance.create_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELCACHE,
                localmodelcache,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->create_cluster_custom_object:{e}%s\n"
            ) from e
        return output

    def get_local_model_cache(
        self, name: str, version: str = constants.KSERVE_V1ALPHA1_VERSION
    ) -> object:
        """
        Get the local model cache

        :param name: existing local model cache name
        :param version: api group version. Default to v1alpha1
        :return: local model cache object
        """
        try:
            return self.api_instance.get_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELCACHE,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->get_cluster_custom_object:{e}%s\n"
            ) from e

    def delete_local_model_cache(
        self, name: str, version: str = constants.KSERVE_V1ALPHA1_VERSION
    ):
        """
        Delete the local model cache

        :param name: local model cache name
        :param version: api group version. Default to v1alpha1
        """
        try:
            self.api_instance.delete_cluster_custom_object(
                constants.KSERVE_GROUP,
                version,
                constants.KSERVE_PLURAL_LOCALMODELCACHE,
                name,
            )
        except client.rest.ApiException as e:
            raise RuntimeError(
                f"Exception when calling CustomObjectsApi->delete_cluster_custom_object:{e}%s\n"
            ) from e

    def is_local_model_cache_ready(
        self,
        name: str,
        nodes: List[str],
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
    ) -> bool:
        """
        Verify if the model is successfully cached on the specified node.

        :param name: local model cache name
        :param node_name: name of the node to check if the model is cached
        :param version: api group version
        :return: true if the model is successfully cached, else false.
        """
        local_model_cache = self.get_local_model_cache(name, version=version)
        node_status = local_model_cache.get("status", {}).get("nodeStatus", {})
        for node in nodes:
            if node_status.get(node, "") != "NodeDownloaded":
                return False
        return True

    def wait_local_model_cache_ready(
        self,
        name: str,
        nodes: List[str],
        version: str = constants.KSERVE_V1ALPHA1_VERSION,
        timeout_seconds: int = 600,
        polling_interval: int = 10,
    ):
        """
        Wait for model to be cached locally for specified nodes until timeout.

        :param name: local model cache name
        :param nodes: list of node names to check if the model is cached
        :param version: api group version
        :param timeout_seconds: timeout seconds for waiting, default to 600s.
        :param polling_interval: The time interval to poll status
        :return:
        """
        for _ in range(round(timeout_seconds / polling_interval)):
            time.sleep(polling_interval)
            if self.is_local_model_cache_ready(name, nodes, version=version):
                return

        current_local_model_cache = self.get_local_model_cache(name, version=version)
        raise RuntimeError(
            f"Timeout while caching the model. The current state of LocalModelCache is: {current_local_model_cache}"
        )



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import AsyncIterator, Iterable, Union, cast, Optional

from kserve.protocol.rest.openai.types import (
    ChatCompletion,
    ChatCompletionChoice,
    ChatCompletionChoiceLogprobs,
    ChatCompletionChunk,
    ChatCompletionRequestMessage,
    ChatCompletionResponseMessage,
    ChatCompletionTokenLogprob,
    ChoiceDelta,
    ChunkChoice,
    Completion,
    CompletionChoice,
    CreateChatCompletionRequest,
    CreateCompletionRequest,
    Logprobs,
    TopLogprob,
)

from ....errors import InvalidInput
from .openai_model import (
    OpenAICompletionModel,
    ChatPrompt,
    CompletionRequest,
    ChatCompletionRequest,
    AsyncMappingIterator,
)


class OpenAIChatAdapterModel(OpenAICompletionModel):
    """
    A helper on top the OpenAI model that automatically maps chat completion requests (/v1/chat/completions)
    to completion requests (/v1/completions).

    Users should extend this model and implement the abstract methods in order to expose these endpoints.
    """

    @abstractmethod
    def apply_chat_template(
        self,
        messages: Iterable[ChatCompletionRequestMessage],
        chat_template: Optional[str] = None,
    ) -> ChatPrompt:
        """
        Given a list of chat completion messages, convert them to a prompt.
        """
        pass

    @classmethod
    def chat_completion_params_to_completion_params(
        cls, params: CreateChatCompletionRequest, prompt: str
    ) -> CreateCompletionRequest:

        return CreateCompletionRequest(
            prompt=prompt,
            model=params.model,
            frequency_penalty=params.frequency_penalty,
            logit_bias=params.logit_bias,
            max_tokens=params.max_tokens,
            n=params.n,
            presence_penalty=params.presence_penalty,
            seed=params.seed,
            stop=params.stop,
            stream=params.stream,
            temperature=params.temperature,
            top_p=params.top_p,
            user=params.user,
            logprobs=params.top_logprobs,
        )

    @classmethod
    def to_choice_logprobs(cls, logprobs: Logprobs) -> ChatCompletionChoiceLogprobs:
        chat_completion_logprobs = []
        for i in range(len(logprobs.tokens)):
            token = logprobs.tokens[i]
            token_logprob = logprobs.token_logprobs[i]
            top_logprobs_dict = logprobs.top_logprobs[i]
            top_logprobs = [
                TopLogprob(
                    token=token,
                    bytes=[int(b) for b in token.encode("utf8")],
                    logprob=logprob,
                )
                for token, logprob in top_logprobs_dict.items()
            ]
            chat_completion_logprobs.append(
                ChatCompletionTokenLogprob(
                    token=token,
                    bytes=[int(b) for b in token.encode("utf8")],
                    logprob=token_logprob,
                    top_logprobs=top_logprobs,
                )
            )

        return ChatCompletionChoiceLogprobs(content=chat_completion_logprobs)

    @classmethod
    def to_chat_completion_choice(
        cls, completion_choice: CompletionChoice, role: str
    ) -> ChatCompletionChoice:
        # translate Token -> ChatCompletionTokenLogprob
        choice_logprobs = (
            cls.to_choice_logprobs(completion_choice.logprobs)
            if completion_choice.logprobs is not None
            else None
        )
        return ChatCompletionChoice(
            index=0,
            finish_reason=completion_choice.finish_reason,
            logprobs=choice_logprobs,
            message=ChatCompletionResponseMessage(
                content=completion_choice.text, role=role
            ),
        )

    @classmethod
    def to_chat_completion_chunk_choice(
        cls, completion_choice: CompletionChoice, role: str
    ) -> ChunkChoice:
        # translate Token -> ChatCompletionTokenLogprob
        choice_logprobs = (
            cls.to_choice_logprobs(completion_choice.logprobs)
            if completion_choice.logprobs is not None
            else None
        )
        choice_logprobs = (
            ChatCompletionChoiceLogprobs(content=choice_logprobs.content)
            if choice_logprobs is not None
            else None
        )
        return ChunkChoice(
            delta=ChoiceDelta(content=completion_choice.text, role=role),
            index=0,
            finish_reason=completion_choice.finish_reason,
            logprobs=choice_logprobs,
        )

    @classmethod
    def completion_to_chat_completion(
        cls, completion: Completion, role: str
    ) -> ChatCompletion:
        completion_choice = (
            completion.choices[0] if len(completion.choices) > 0 else None
        )
        choices = (
            [cls.to_chat_completion_choice(completion_choice, role)]
            if completion_choice is not None
            else []
        )
        return ChatCompletion(
            id=completion.id,
            choices=choices,
            created=completion.created,
            model=completion.model,
            object="chat.completion",
            system_fingerprint=completion.system_fingerprint,
            usage=completion.usage,
        )

    @classmethod
    def completion_to_chat_completion_chunk(
        cls, completion: Completion, role: str
    ) -> ChatCompletionChunk:
        completion_choice = (
            completion.choices[0] if len(completion.choices) > 0 else None
        )
        choices = (
            [cls.to_chat_completion_chunk_choice(completion_choice, role)]
            if completion_choice is not None
            else []
        )
        return ChatCompletionChunk(
            id=completion.id,
            choices=choices,
            created=completion.created,
            model=completion.model,
            object="chat.completion.chunk",
            system_fingerprint=completion.system_fingerprint,
        )

    async def create_chat_completion(
        self,
        request: ChatCompletionRequest,
    ) -> Union[ChatCompletion, AsyncIterator[ChatCompletionChunk]]:
        params = request.params

        if params.n != 1:
            raise InvalidInput("n != 1 is not supported")

        # Convert the messages into a prompt
        chat_prompt = self.apply_chat_template(
            params.messages, params.chat_template, params.tools
        )
        # Translate the chat completion request to a completion request
        completion_params = self.chat_completion_params_to_completion_params(
            params, chat_prompt.prompt
        )

        completion_request = CompletionRequest(
            request_id=request.request_id,
            params=completion_params,
            context=request.context,
        )

        if not params.stream:
            completion = cast(
                Completion, await self.create_completion(completion_request)
            )
            return self.completion_to_chat_completion(
                completion, chat_prompt.response_role
            )
        else:
            completion_iterator = cast(
                AsyncIterator[Completion],
                await self.create_completion(completion_request),
            )

            def mapper(completion: Completion) -> ChatCompletionChunk:
                return self.completion_to_chat_completion_chunk(
                    completion, chat_prompt.response_role
                )

            return AsyncMappingIterator(iterator=completion_iterator, mapper=mapper)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, Optional

from .model import BaseKServeModel

MODEL_MOUNT_DIRS = "/mnt/models"


class ModelRepository:
    """Model repository interface.

    It follows NVIDIA Triton's `Model Repository Extension`_.

    .. _Model Repository Extension:
        https://github.com/triton-inference-server/server/blob/main/docs/protocol/extension_model_repository.md
    """

    def __init__(self, models_dir: str = MODEL_MOUNT_DIRS):
        self.models: Dict[str, BaseKServeModel] = {}
        self.models_dir = models_dir

    def load_models(self):
        for name in os.listdir(self.models_dir):
            d = os.path.join(self.models_dir, name)
            if os.path.isdir(d):
                self.load_model(name)

    def set_models_dir(self, models_dir):  # used for unit tests
        self.models_dir = models_dir

    def get_model(self, name: str) -> Optional[BaseKServeModel]:
        return self.models.get(name, None)

    def get_models(self) -> Dict[str, BaseKServeModel]:
        return self.models

    async def is_model_ready(self, name: str):
        model = self.get_model(name)
        if not model:
            return False
        return await model.healthy()

    def update(self, model: BaseKServeModel):
        self.models[model.name] = model

    def load(self, name: str) -> bool:
        pass

    def load_model(self, name: str) -> bool:
        pass

    def unload(self, name: str):
        if name in self.models:
            model = self.models[name]
            if callable(getattr(model, "stop", None)):
                model.stop()
            del self.models[name]
        else:
            raise KeyError(f"model with name {name} does not exist")



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import configparser
from os.path import expanduser

from kubernetes import client
from ..constants import constants

from kserve.logging import logger


def set_gcs_credentials(namespace, credentials_file, service_account):
    """
    Set GCS Credentials (secret and service account) with credentials file.
    Args:
        namespace(str): The kubernetes namespace.
        credentials_file(str): The path for the gcs credentials file.
        service_account(str): The name of service account. If the service_account
                              is specified, will attach created secret with the service account,
                              otherwise will create new one and attach with created secret.
    """

    with open(expanduser(credentials_file)) as f:
        gcs_creds_content = f.read()

    # Try to get GCS creds file name from configmap, set default value then if cannot.
    gcs_creds_file_name = get_creds_name_from_config_map("gcsCredentialFileName")
    if not gcs_creds_file_name:
        gcs_creds_file_name = constants.GCS_CREDS_FILE_DEFAULT_NAME

    string_data = {gcs_creds_file_name: gcs_creds_content}

    secret_name = create_secret(namespace=namespace, string_data=string_data)

    set_service_account(
        namespace=namespace, service_account=service_account, secret_name=secret_name
    )


def set_s3_credentials(
    namespace,
    credentials_file,
    service_account,
    s3_profile="default",  # pylint: disable=too-many-locals,too-many-arguments
    s3_endpoint=None,
    s3_region=None,
    s3_use_https=None,
    s3_verify_ssl=None,
    s3_cabundle=None,
):  # pylint: disable=unused-argument
    """
    Set S3 Credentials (secret and service account).
    Args:
        namespace(str): The kubernetes namespace.
        credentials_file(str): The path for the S3 credentials file.
        s3_profile(str): The profile for S3, default value is 'default'.
        service_account(str): The name of service account(Optional). If the service_account
                              is specified, will attach created secret with the service account,
                              otherwise will create new one and attach with created secret.
        s3_endpoint(str): S3 settings variable S3_ENDPOINT.
        s3_region(str): S3 settings variable AWS_DEFAULT_REGION.
        s3_use_https(str): S3 settings variable S3_USE_HTTPS.
        s3_verify_ssl(str): S3 settings variable S3_VERIFY_SSL.
        s3_cabundle(str): S3 settings variable AWS_CA_BUNDLE.
    """

    config = configparser.ConfigParser()
    config.read([expanduser(credentials_file)])
    s3_access_key_id = config.get(s3_profile, "aws_access_key_id")
    s3_secret_access_key = config.get(s3_profile, "aws_secret_access_key")

    # Try to get S3 creds name from configmap, set default value then if cannot.
    s3_access_key_id_name = get_creds_name_from_config_map("s3AccessKeyIDName")
    if not s3_access_key_id_name:
        s3_access_key_id_name = constants.S3_ACCESS_KEY_ID_DEFAULT_NAME

    s3_secret_access_key_name = get_creds_name_from_config_map("s3SecretAccessKeyName")
    if not s3_secret_access_key_name:
        s3_secret_access_key_name = constants.S3_SECRET_ACCESS_KEY_DEFAULT_NAME

    data = {
        s3_access_key_id_name: s3_access_key_id,
        s3_secret_access_key_name: s3_secret_access_key,
    }

    s3_cred_sets = {
        "s3_endpoint": constants.KSERVE_GROUP + "/s3-endpoint",
        "s3_region": constants.KSERVE_GROUP + "/s3-region",
        "s3_use_https": constants.KSERVE_GROUP + "/s3-usehttps",
        "s3_verify_ssl": constants.KSERVE_GROUP + "/s3-verifyssl",
        "s3_cabundle": constants.KSERVE_GROUP + "/s3-cabundle",
    }

    s3_annotations = {}
    for key, value in s3_cred_sets.items():
        arg = vars()[key]
        if arg is not None:
            s3_annotations.update({value: arg})

    secret_name = create_secret(
        namespace=namespace, annotations=s3_annotations, data=data
    )

    set_service_account(
        namespace=namespace, service_account=service_account, secret_name=secret_name
    )


def set_azure_credentials(namespace, credentials_file, service_account):
    """
    Set Azure Credentials (secret and service account) with credentials file.
    Args:
        namespace(str): The kubernetes namespace.
        credentials_file(str): The path for the Azure credentials file.
        service_account(str): The name of service account. If the service_account
                              is specified, will attach created secret with the service account,
                              otherwise will create new one and attach with created secret.
    """

    with open(expanduser(credentials_file)) as azure_creds_file:
        azure_creds = json.load(azure_creds_file)

    data = {
        "AZURE_CLIENT_ID": azure_creds["clientId"],
        "AZURE_CLIENT_SECRET": azure_creds["clientSecret"],
        "AZURE_SUBSCRIPTION_ID": azure_creds["subscriptionId"],
        "AZURE_TENANT_ID": azure_creds["tenantId"],
    }

    secret_name = create_secret(namespace=namespace, data=data)

    set_service_account(
        namespace=namespace, service_account=service_account, secret_name=secret_name
    )


def create_secret(namespace, annotations=None, data=None, string_data=None):
    "Create namespaced secret, and return the secret name."
    try:
        created_secret = client.CoreV1Api().create_namespaced_secret(
            namespace,
            client.V1Secret(
                api_version="v1",
                kind="Secret",
                metadata=client.V1ObjectMeta(
                    generate_name=constants.DEFAULT_SECRET_NAME, annotations=annotations
                ),
                data=data,
                string_data=string_data,
            ),
        )
    except client.rest.ApiException as e:
        raise RuntimeError(
            "Exception when calling CoreV1Api->create_namespaced_secret: %s\n" % e
        )

    secret_name = created_secret.metadata.name
    logger.info("Created Secret: %s in namespace %s", secret_name, namespace)
    return secret_name


def set_service_account(namespace, service_account, secret_name):
    """
    Set service account, create if service_account does not exist, otherwise patch it.
    """
    if check_sa_exists(namespace=namespace, service_account=service_account):
        patch_service_account(
            secret_name=secret_name, namespace=namespace, sa_name=service_account
        )
    else:
        create_service_account(
            secret_name=secret_name, namespace=namespace, sa_name=service_account
        )


def check_sa_exists(namespace, service_account):
    """
    Check if the specified service account existing.
    """
    sa_list = client.CoreV1Api().list_namespaced_service_account(namespace=namespace)

    sa_name_list = [sa.metadata.name for sa in sa_list.items]

    if service_account in sa_name_list:
        return True

    return False


def create_service_account(secret_name, namespace, sa_name):
    """
    Create namespaced service account, and return the service account name
    """
    try:
        client.CoreV1Api().create_namespaced_service_account(
            namespace,
            client.V1ServiceAccount(
                metadata=client.V1ObjectMeta(name=sa_name),
                secrets=[client.V1ObjectReference(kind="Secret", name=secret_name)],
            ),
        )
    except client.rest.ApiException as e:
        raise RuntimeError(
            "Exception when calling CoreV1Api->create_namespaced_service_account: %s\n"
            % e
        )

    logger.info("Created Service account: %s in namespace %s", sa_name, namespace)


def patch_service_account(secret_name, namespace, sa_name):
    """
    Patch namespaced service account to attach with created secret.
    """
    try:
        client.CoreV1Api().patch_namespaced_service_account(
            sa_name,
            namespace,
            client.V1ServiceAccount(
                secrets=[client.V1ObjectReference(kind="Secret", name=secret_name)]
            ),
        )
    except client.rest.ApiException as e:
        raise RuntimeError(
            "Exception when calling CoreV1Api->patch_namespaced_service_account: %s\n"
            % e
        )

    logger.info("Patched Service account: %s in namespace %s", sa_name, namespace)


def get_creds_name_from_config_map(creds):
    """
    Get the credentials name from inferenceservice config map.
    """
    try:
        isvc_config_map = client.CoreV1Api().read_namespaced_config_map(
            constants.INFERENCESERVICE_CONFIG_MAP_NAME,
            constants.INFERENCESERVICE_SYSTEM_NAMESPACE,
        )
    except client.rest.ApiException:
        logger.warning(
            "Cannot get configmap %s in namespace %s.",
            constants.INFERENCESERVICE_CONFIG_MAP_NAME,
            constants.INFERENCESERVICE_SYSTEM_NAMESPACE,
        )
        return None

    isvc_creds_str = isvc_config_map.data["credentials"]
    isvc_creds_json = json.loads(isvc_creds_str)

    if creds == "gcsCredentialFileName":
        return isvc_creds_json["gcs"]["gcsCredentialFileName"]
    elif creds == "s3AccessKeyIDName":
        return isvc_creds_json["s3"]["s3AccessKeyIDName"]
    elif creds == "s3SecretAccessKeyName":
        return isvc_creds_json["s3"]["s3SecretAccessKeyName"]
    else:
        raise RuntimeError("Unknown credentials.")



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import concurrent.futures
import signal
import sys
from importlib import metadata
from typing import Any, Callable, Dict, List, Optional

from fastapi import FastAPI
from fastapi.responses import ORJSONResponse

from . import logging
from .constants.constants import (
    DEFAULT_HTTP_PORT,
    DEFAULT_GRPC_PORT,
    MAX_GRPC_MESSAGE_LENGTH,
)
from .errors import NoModelReady
from .logging import logger
from .model import BaseKServeModel, PredictorConfig
from .model_repository import ModelRepository
from .protocol.dataplane import DataPlane
from .protocol.grpc.server import GRPCServer
from .protocol.model_repository_extension import ModelRepositoryExtension
from .protocol.rest.server import RESTServer
from .utils import utils
from .utils.inference_client_factory import InferenceClientFactory


parser = argparse.ArgumentParser(
    add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
# Model Server Arguments: The arguments are passed to the kserve.ModelServer object
parser.add_argument(
    "--http_port",
    default=DEFAULT_HTTP_PORT,
    type=int,
    help="The HTTP Port listened to by the model server.",
)
parser.add_argument(
    "--grpc_port",
    default=DEFAULT_GRPC_PORT,
    type=int,
    help="The GRPC Port listened to by the model server.",
)
parser.add_argument(
    "--workers",
    default=1,
    type=int,
    help="The number of uvicorn workers for multi-processing.",
)
parser.add_argument(
    "--max_threads",
    default=4,
    type=int,
    help="The max number of gRPC processing threads.",
)
parser.add_argument(
    "--max_asyncio_workers",
    default=None,
    type=int,
    help="The max number of asyncio workers to spawn.",
)
parser.add_argument(
    "--enable_grpc",
    default=True,
    type=lambda x: utils.strtobool(x),
    help="Enable gRPC for the model server.",
)
parser.add_argument(
    "--enable_docs_url",
    default=False,
    type=lambda x: utils.strtobool(x),
    help="Enable docs url '/docs' to display Swagger UI.",
)
parser.add_argument(
    "--enable_latency_logging",
    default=True,
    type=lambda x: utils.strtobool(x),
    help="Enable a log line per request with preprocess/predict/postprocess latency metrics.",
)
parser.add_argument(
    "--configure_logging",
    default=True,
    type=lambda x: utils.strtobool(x),
    help="Enable to configure KServe and Uvicorn logging.",
)
parser.add_argument(
    "--log_config_file",
    default=None,
    type=str,
    help="File path containing UvicornServer's log config. Needs to be a yaml or json file.",
)
parser.add_argument(
    "--access_log_format",
    default=None,
    type=str,
    help="The asgi access logging format. It allows to override only the `uvicorn.access`'s format configuration "
    "with a richer set of fields",
)

# Model arguments: The arguments are passed to the kserve.Model object
parser.add_argument(
    "--model_name",
    default="model",
    type=str,
    help="The name of the model used on the endpoint path.",
)
parser.add_argument(
    "--predictor_host",
    default=None,
    type=str,
    help="The host name used for calling to the predictor from transformer.",
)
# For backwards compatibility.
parser.add_argument(
    "--protocol",
    default="v1",
    type=str,
    choices=["v1", "v2", "grpc-v2"],
    help="The inference protocol used for calling to the predictor from transformer. "
    "Deprecated and replaced by --predictor_protocol",
)
parser.add_argument(
    "--predictor_protocol",
    default="v1",
    type=str,
    choices=["v1", "v2", "grpc-v2"],
    help="The inference protocol used for calling to the predictor from transformer.",
)
parser.add_argument(
    "--predictor_use_ssl",
    default=False,
    type=lambda x: utils.strtobool(x),
    help="Use ssl for the http connection to the predictor.",
)
parser.add_argument(
    "--predictor_request_timeout_seconds",
    default=600,
    type=int,
    help="The timeout seconds for the request sent to the predictor.",
)
parser.add_argument(
    "--grpc_max_send_message_length",
    default=MAX_GRPC_MESSAGE_LENGTH,
    type=int,
    help="The max message length for gRPC send message.",
)
parser.add_argument(
    "--grpc_max_receive_message_length",
    default=MAX_GRPC_MESSAGE_LENGTH,
    type=int,
    help="The max message length for gRPC receive message.",
)
parser.add_argument(
    "--predictor_request_retries",
    default=0,
    type=int,
    help="The number of retries if predictor request fails. Defaults to 0.",
)
parser.add_argument(
    "--enable_predictor_health_check",
    action="store_true",
    help="The Transformer will perform readiness check for the predictor in addition to "
    "its health check. By default it is disabled.",
)
args, _ = parser.parse_known_args()

app = FastAPI(
    title="KServe ModelServer",
    version=metadata.version("kserve"),
    docs_url="/docs" if args.enable_docs_url else None,
    redoc_url=None,
    default_response_class=ORJSONResponse,
)


class ModelServer:
    def __init__(
        self,
        http_port: int = args.http_port,
        grpc_port: int = args.grpc_port,
        workers: int = args.workers,
        max_threads: int = args.max_threads,
        max_asyncio_workers: int = args.max_asyncio_workers,
        registered_models: Optional[ModelRepository] = None,
        enable_grpc: bool = args.enable_grpc,
        enable_docs_url: bool = args.enable_docs_url,
        enable_latency_logging: bool = args.enable_latency_logging,
        access_log_format: str = args.access_log_format,
    ):
        """KServe ModelServer Constructor

        Args:
            http_port: HTTP port. Default: ``8080``.
            grpc_port: GRPC port. Default: ``8081``.
            workers: Number of uvicorn workers. Default: ``1``.
            max_threads: Max number of gRPC processing threads. Default: ``4``
            max_asyncio_workers: Max number of AsyncIO threads. Default: ``None``
            registered_models: A optional Model repository with registered models.
            enable_grpc: Whether to turn on grpc server. Default: ``True``
            enable_docs_url: Whether to turn on ``/docs`` Swagger UI. Default: ``False``.
            enable_latency_logging: Whether to log latency metric. Default: ``True``.
            access_log_format: Format to set for the access log (provided by asgi-logger). Default: ``None``.
                               it allows to override only the `uvicorn.access`'s format configuration with a richer
                               set of fields (output hardcoded to `stdout`). This limitation is currently due to the
                               ASGI specs that don't describe how access logging should be implemented in detail
                               (please refer to this Uvicorn
                               [github issue](https://github.com/encode/uvicorn/issues/527) for more info).
        """
        self.registered_models = (
            ModelRepository() if registered_models is None else registered_models
        )
        self.http_port = http_port
        self.grpc_port = grpc_port
        self.workers = workers
        self.max_threads = max_threads
        self.max_asyncio_workers = max_asyncio_workers
        self.enable_grpc = enable_grpc
        self.enable_docs_url = enable_docs_url
        self.enable_latency_logging = enable_latency_logging
        self.dataplane = DataPlane(model_registry=self.registered_models)
        self.model_repository_extension = ModelRepositoryExtension(
            model_registry=self.registered_models
        )
        if args.configure_logging:
            # If the logger does not have any handlers, then the logger is not configured.
            # For backward compatibility, we configure the logger here.
            if len(logger.handlers) == 0:
                logging.configure_logging(args.log_config_file)
        self.access_log_format = access_log_format
        self._custom_exception_handler = None
        predictor_config = PredictorConfig(
            predictor_host=args.predictor_host,
            predictor_protocol=args.predictor_protocol,
            predictor_use_ssl=args.predictor_use_ssl,
            predictor_request_timeout_seconds=args.predictor_request_timeout_seconds,
            predictor_request_retries=args.predictor_request_retries,
            predictor_health_check=args.enable_predictor_health_check,
        )
        self.dataplane = DataPlane(
            model_registry=self.registered_models, predictor_config=predictor_config
        )
        self._rest_server = self._rest_server = RESTServer(
            app,
            self.dataplane,
            self.model_repository_extension,
            self.http_port,
            # By setting log_config to None we tell Uvicorn not to configure logging as it is already
            # configured by kserve.
            log_config=None,
            access_log_format=self.access_log_format,
            workers=self.workers,
        )
        self._grpc_server = None
        if self.enable_grpc:
            self._grpc_server = GRPCServer(
                grpc_port,
                self.dataplane,
                self.model_repository_extension,
                kwargs=vars(args),
            )

    def setup_event_loop(self):
        loop = asyncio.get_event_loop()
        if self._custom_exception_handler is None:
            loop.set_exception_handler(self.default_exception_handler)
        else:
            loop.set_exception_handler(self._custom_exception_handler)

        if self.max_asyncio_workers is None:
            # formula as suggest in https://bugs.python.org/issue35279
            self.max_asyncio_workers = min(32, utils.cpu_count() + 4)
        logger.info(f"Setting max asyncio worker threads as {self.max_asyncio_workers}")
        loop.set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self.max_asyncio_workers)
        )

    def register_signal_handler(self):
        if sys.platform == "win32":
            sig_list = [signal.SIGINT, signal.SIGTERM, signal.SIGBREAK]
        else:
            sig_list = [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]

        for sig in sig_list:
            signal.signal(
                sig, lambda sig, frame: asyncio.create_task(self.stop(sig=sig))
            )

    async def _servers_task(self):
        servers = [self._rest_server.start()]
        if self.enable_grpc:
            servers.append(self._grpc_server.start(self.max_threads))
        await asyncio.gather(*servers)

    def start(self, models: List[BaseKServeModel]) -> None:
        """Start the model server with a set of registered models.

        Args:
            models: a list of models to register to the model server.
        """
        self._check_atleast_one_model_is_ready(models)
        self.setup_event_loop()
        self.register_signal_handler()
        asyncio.run(self._servers_task())

    async def stop(self, sig: int):
        """Stop the instances of REST and gRPC model servers.

        Args:
            sig: The signal to stop the server.
        """
        await InferenceClientFactory().close()
        logger.info("Stopping the model server")
        if self._grpc_server:
            logger.info("Stopping the grpc server")
            await self._grpc_server.stop(sig)
        for model_name in list(self.registered_models.get_models().keys()):
            self.registered_models.unload(model_name)

    def register_exception_handler(
        self,
        handler: Callable[[asyncio.events.AbstractEventLoop, Dict[str, Any]], None],
    ):
        """Add a custom handler as the event loop exception handler.

        If a handler is not provided, the default exception handler will be set.

        handler should be a callable object, it should have a signature matching '(loop, context)', where 'loop'
        will be a reference to the active event loop, 'context' will be a dict object (see `call_exception_handler()`
        documentation for details about context).
        """
        self._custom_exception_handler = handler

    def default_exception_handler(
        self, loop: asyncio.events.AbstractEventLoop, context: Dict[str, Any]
    ):
        """Default exception handler for event loop.

        This is called when an exception occurs and no exception handler is set.
        This can be called by a custom exception handler that wants to defer to the default handler behavior.
        """
        if "exception" in context:
            logger.error(f"Caught exception: {context.get('exception')}")
        logger.error(f"message: {context.get('message')}")
        loop.default_exception_handler(context)

    def register_model(self, model: BaseKServeModel):
        """Register a model to the model server.

        Args:
            model: The model object.
        """
        if not model.name:
            raise Exception("Failed to register model, model.name must be provided.")
        self.registered_models.update(model)
        logger.info("Registering model: %s", model.name)

    def _check_atleast_one_model_is_ready(self, models: List[BaseKServeModel]):
        if isinstance(models, list):
            at_least_one_model_ready = False
            for model in models:
                if isinstance(model, BaseKServeModel):
                    if model.ready:
                        at_least_one_model_ready = True
                        self.register_model(model)
                        # pass whether to log request latency into the model
                        model.enable_latency_logging = self.enable_latency_logging
                    model.start()
                else:
                    raise RuntimeError("Model type should be 'BaseKServeModel'")
            if not at_least_one_model_ready and models:
                raise NoModelReady(models)
        else:
            raise RuntimeError("Unknown model collection type")



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List

from kserve.protocol.rest.openai.types.openapi import (
    ChatCompletionRequestAssistantMessage,
    ChatCompletionRequestFunctionMessage,
    ChatCompletionRequestSystemMessage,
    ChatCompletionRequestToolMessage,
    ChatCompletionRequestUserMessage,
    ChatCompletionResponseMessage,
)
from kserve.protocol.rest.openai.types.openapi import (
    ChatCompletionStreamResponseDelta as ChoiceDelta,
)
from kserve.protocol.rest.openai.types.openapi import ChatCompletionTokenLogprob
from kserve.protocol.rest.openai.types.openapi import Choice as CompletionChoice
from kserve.protocol.rest.openai.types.openapi import Choice1 as ChatCompletionChoice
from kserve.protocol.rest.openai.types.openapi import Choice3 as ChunkChoice
from kserve.protocol.rest.openai.types.openapi import CreateChatCompletionRequest
from kserve.protocol.rest.openai.types.openapi import (
    CreateChatCompletionResponse as ChatCompletion,
)
from kserve.protocol.rest.openai.types.openapi import (
    CreateChatCompletionStreamResponse as ChatCompletionChunk,
)
from kserve.protocol.rest.openai.types.openapi import CreateCompletionRequest
from kserve.protocol.rest.openai.types.openapi import (
    CreateCompletionResponse as Completion,
)
from kserve.protocol.rest.openai.types.openapi import CreateEmbeddingRequest
from kserve.protocol.rest.openai.types.openapi import (
    CreateEmbeddingResponse as Embedding,
    Embedding as BaseEmbeddingObject,
)
from kserve.protocol.rest.openai.types.openapi import Logprobs
from kserve.protocol.rest.openai.types.openapi import (
    Logprobs2 as ChatCompletionChoiceLogprobs,
)
from kserve.protocol.rest.openai.types.openapi import TopLogprob
from kserve.protocol.rest.openai.types.openapi import ErrorResponse
from kserve.protocol.rest.openai.types.openapi import CompletionUsage

ChatCompletionRequestMessage = Union[
    ChatCompletionRequestSystemMessage,
    ChatCompletionRequestUserMessage,
    ChatCompletionRequestAssistantMessage,
    ChatCompletionRequestToolMessage,
    ChatCompletionRequestFunctionMessage,
]


# The autogenerated OpenAPI specification doesn't seem to include the base64 encoding option. Add it manually.
class EmbeddingObject(BaseEmbeddingObject):
    embedding: Union[List[float], str]


__all__ = [
    "ChatCompletion",
    "ChatCompletionChoice",
    "ChatCompletionChoiceLogprobs",
    "ChatCompletionChoiceLogprobs",
    "ChatCompletionChunk",
    "ChatCompletionRequestAssistantMessage",
    "ChatCompletionRequestFunctionMessage",
    "ChatCompletionRequestMessage",
    "ChatCompletionRequestSystemMessage",
    "ChatCompletionRequestToolMessage",
    "ChatCompletionRequestUserMessage",
    "ChatCompletionResponseMessage",
    "ChatCompletionTokenLogprob",
    "ChoiceDelta",
    "ChunkChoice",
    "Completion",
    "CompletionChoice",
    "CreateChatCompletionRequest",
    "CreateCompletionRequest",
    "CreateEmbeddingRequest",
    "Embedding",
    "EmbeddingObject",
    "ErrorResponse",
    "Logprobs",
    "TopLogprob",
    "CompletionUsage",
]



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Any, AsyncIterator, Callable, Dict, Optional, Union
import inspect

from pydantic import BaseModel

from kserve.protocol.rest.openai.types import (
    ChatCompletion,
    ChatCompletionChunk,
    Completion,
    CreateChatCompletionRequest,
    CreateCompletionRequest,
    CreateEmbeddingRequest,
    Embedding,
)

from ....model import BaseKServeModel


class ChatPrompt(BaseModel):
    response_role: str = "assistant"
    prompt: str


class BaseOpenAIRequest(BaseModel):
    request_id: Optional[str] = None
    context: Optional[Dict[str, Any]] = None  # headers can go in here
    params: Union[
        CreateCompletionRequest, CreateChatCompletionRequest, CreateEmbeddingRequest
    ]


class CompletionRequest(BaseOpenAIRequest):
    params: CreateCompletionRequest


class ChatCompletionRequest(BaseOpenAIRequest):
    params: CreateChatCompletionRequest


class EmbeddingRequest(BaseOpenAIRequest):
    params: CreateEmbeddingRequest


class OpenAIModel(BaseKServeModel):
    """
    An abstract model with methods for implementing OpenAI's endpoints.
    """

    def __init__(self, name: str):
        super().__init__(name)

        # We don't support the `load()` method on OpenAIModel yet
        # Assume the model is ready
        self.ready = True


class OpenAICompletionModel(OpenAIModel):
    """
    An abstract model with methods for implementing OpenAI's completions (v1/completions)
    and chat completions (v1/chat/completions) endpoints.

    Users should extend this model and implement the abstract methods in order to expose
    these endpoints.
    """

    @abstractmethod
    async def create_completion(
        self, request: CompletionRequest
    ) -> Union[Completion, AsyncIterator[Completion]]:
        pass

    @abstractmethod
    async def create_chat_completion(
        self, request: ChatCompletionRequest
    ) -> Union[ChatCompletion, AsyncIterator[ChatCompletionChunk]]:
        pass


class OpenAIEmbeddingModel(OpenAIModel):
    """
    An abstract model with a method for implementing OpenAI's embeddings (/v1/embeddings)
    endpoint.

    Users should extend this model and implement the abstract methods in order to expose
    these endpoints.
    """

    @abstractmethod
    async def create_embedding(self, request: EmbeddingRequest) -> Embedding:
        pass


class AsyncMappingIterator:
    def __init__(
        self,
        iterator: AsyncIterator,
        mapper: Callable = lambda item: item,
        skip_none: bool = True,
        close: Optional[Callable] = None,
    ):
        self.iterator = iterator
        self.mapper = mapper
        self.skip_none = skip_none
        self.close = close

    def __aiter__(self):
        return self

    async def __anext__(self):
        # This will raise StopAsyncIteration when there are no more completions.
        # We don't catch it so it will stop our iterator as well.
        async def next():
            try:
                return self.mapper(await self.iterator.__anext__())
            except Exception as e:
                if self.close:
                    if inspect.iscoroutinefunction(self.close):
                        await self.close()
                    else:
                        self.close()
                raise e

        mapped_item = await next()
        if self.skip_none:
            while mapped_item is None:
                mapped_item = await next()
        return mapped_item



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from importlib import metadata
from typing import Dict, Optional, Tuple, Union, cast
from inspect import iscoroutinefunction

import cloudevents.exceptions as ce
import orjson

from cloudevents.http import CloudEvent, from_http
from cloudevents.sdk.converters.util import has_binary_headers

from ..inference_client import RESTConfig
from .rest.v2_datamodels import InferenceRequest
from ..constants import constants
from ..constants.constants import INFERENCE_CONTENT_LENGTH_HEADER, PredictorProtocol
from ..errors import InvalidInput, ModelNotFound
from ..logging import logger
from ..model import PredictorConfig
from ..model import InferenceVerb, BaseKServeModel, InferenceModel
from ..model_repository import ModelRepository
from ..utils.inference_client_factory import InferenceClientFactory
from ..utils.utils import create_response_cloudevent, is_structured_cloudevent
from .infer_type import InferRequest, InferResponse
from .rest.openai import OpenAICompletionModel

JSON_HEADERS = [
    "application/json",
    "application/cloudevents+json",
    "application/ld+json",
    "application/octet-stream",
]


class DataPlane:
    """KServe DataPlane"""

    def __init__(
        self,
        model_registry: ModelRepository,
        predictor_config: Optional[PredictorConfig] = None,
    ):
        self._model_registry = model_registry
        self._server_name = constants.KSERVE_MODEL_SERVER_NAME

        # Dynamically fetching version of the installed 'kserve' distribution. The assumption is
        # that 'kserve' will already be installed by the time this class is instantiated.
        self._server_version = metadata.version("kserve")
        self.predictor_config = predictor_config
        self._inference_grpc_client = None
        self._inference_rest_client = None

    @property
    def model_registry(self):
        return self._model_registry

    @property
    def rest_client(self):
        if self._inference_rest_client is None:
            self._inference_rest_client = InferenceClientFactory().get_rest_client(
                RESTConfig(
                    protocol=self.predictor_config.predictor_protocol,
                    retries=self.predictor_config.predictor_request_retries,
                    timeout=self.predictor_config.predictor_request_timeout_seconds,
                )
            )
        return self._inference_rest_client

    @property
    def grpc_client(self):
        if self._inference_grpc_client is None:
            self._inference_grpc_client = InferenceClientFactory().get_grpc_client(
                url=self.predictor_config.predictor_host,
                timeout=self.predictor_config.predictor_request_timeout_seconds,
                retries=self.predictor_config.predictor_request_retries,
                use_ssl=self.predictor_config.predictor_use_ssl,
            )
        return self._inference_grpc_client

    def get_model_from_registry(self, name: str) -> BaseKServeModel:
        model = self._model_registry.get_model(name)
        if model is None:
            raise ModelNotFound(name)

        return model

    async def get_model(self, name: str) -> BaseKServeModel:
        """Get the model instance with the given name.

        Args:
            name (str): Model name.

        Returns:
            ModelHandleType: Instance of the model.
        """
        model = self._model_registry.get_model(name)
        if model is None:
            raise ModelNotFound(name)
        is_ready = await self._model_registry.is_model_ready(name)
        if not is_ready:
            model.load()
        return model

    @staticmethod
    def get_binary_cloudevent(
        body: Union[str, bytes, None], headers: Dict[str, str]
    ) -> CloudEvent:
        """Helper function to parse CloudEvent body and headers.

        Args:
            body (str|bytes|None): Request body.
            headers (Dict[str, str]): Request headers.

        Returns:
            CloudEvent: A CloudEvent instance parsed from http body and headers.

        Raises:
            InvalidInput: An error when CloudEvent failed to parse.
        """
        try:
            # Use default unmarshaller if contenttype is set in header
            if "ce-contenttype" in headers:
                event = from_http(headers, body)
            else:
                event = from_http(headers, body, lambda x: x)

            return event
        except (
            ce.MissingRequiredFields,
            ce.InvalidRequiredFields,
            ce.InvalidStructuredJSON,
            ce.InvalidHeadersFormat,
            ce.DataMarshallerError,
            ce.DataUnmarshallerError,
        ) as e:
            raise InvalidInput(f"Cloud Event Exceptions: {e}")

    @staticmethod
    async def live() -> Dict[str, str]:
        """Server live.

        Returns ``{"status": "alive"}`` on successful invocation.
        Primarily meant to be used for Kubernetes liveness check.

        Returns:
            Dict: {"status": "alive"}
        """
        return {"status": "alive"}

    def metadata(self) -> Dict:
        """Server metadata.

        Note:
            Supports ``model_repository_extension`` as defined at Triton Server `Model Repository Extension`_.

        Returns:
            Returns a dict object with following fields:
                - name (str): name of the server.
                - version (str): server version number.
                - extension (list[str]): list of protocol extensions supported by this server.

        .. _Model Repository Extension:
            https://github.com/triton-inference-server/server/blob/main/docs/protocol/extension_model_repository.md
        """
        return {
            "name": self._server_name,
            "version": self._server_version,
            "extensions": ["model_repository_extension"],
        }

    async def model_metadata(self, model_name: str) -> Dict:
        """Get metadata for a specific model.

        Args:
            model_name (str): Model name

        Returns:
            Dict: dictionary with following fields:

                - name (str): name of the model
                - platform: "" (Empty String)
                - inputs: Dict with below fields
                    - name (str): name of the input
                    - datatype (str): Eg. INT32, FP32
                    - shape ([]int): The shape of the tensor.
                                   Variable-size dimensions are specified as -1.
                - outputs: Same as inputs described above.

        Note:
            For more information, check KServe v2 `Model Metadata`_ documentation.

        Raises:
            ModelNotFound: exception will be raised if the model with model_name is not found.

        .. _Model Metadata:
            https://github.com/kserve/kserve/blob/master/docs/predict-api/v2/required_api.md#model-metadata
        """
        # TODO: model versioning is not supported yet
        model = self.get_model_from_registry(model_name)

        if not isinstance(model, InferenceModel):
            raise ValueError(
                f"Model of type {type(model).__name__} does not support inference"
            )
        input_types = (
            await model.get_input_types()
            if iscoroutinefunction(model.get_input_types)
            else model.get_input_types()
        )
        output_types = (
            await model.get_output_types()
            if iscoroutinefunction(model.get_output_types)
            else model.get_output_types()
        )

        return {
            "name": model_name,
            "platform": "",
            "inputs": input_types,
            "outputs": output_types,
        }

    async def ready(self) -> bool:
        """Server ready.

        Returns ``True``. Primarily meant to be used as Kubernetes readiness check.

        Returns:
            bool: True
        """
        # If predictor host is present, then it means this is a transformer,
        # We should also need to check the predictor server's health if predictor health check is enabled.
        if self.predictor_config and self.predictor_config.predictor_health_check:
            if (
                self.predictor_config.predictor_protocol
                == PredictorProtocol.GRPC_V2.value
            ):
                return await self.grpc_client.is_server_ready()
            elif (
                self.predictor_config.predictor_protocol
                == PredictorProtocol.REST_V1.value
            ):
                # V1 Protocol does not have readiness endpoint. We will use server liveness endpoint instead.
                return await self.rest_client.is_server_live(
                    self.predictor_config.predictor_base_url
                )
            else:
                return await self.rest_client.is_server_ready(
                    self.predictor_config.predictor_base_url
                )
        return True

    async def model_ready(self, model_name: str) -> bool:
        """Check if a model is ready.

        Args:
            model_name (str): name of the model

        Returns:
            bool: True if the model is ready, False otherwise.

        Raises:
            ModelNotFound: exception if model is not found
        """
        if self._model_registry.get_model(model_name) is None:
            raise ModelNotFound(model_name)

        # If predictor host is present, then it means this is a transformer,
        # We should also check the predictor model's health if predictor health check is enabled.
        if self.predictor_config and self.predictor_config.predictor_health_check:
            if (
                self.predictor_config.predictor_protocol
                == PredictorProtocol.GRPC_V2.value
            ):
                is_ready = await self.grpc_client.is_model_ready(model_name=model_name)
                return is_ready
            else:
                is_ready = await self.rest_client.is_model_ready(
                    base_url=self.predictor_config.predictor_base_url,
                    model_name=model_name,
                )
                return is_ready

        return await self._model_registry.is_model_ready(model_name)

    def decode(
        self,
        body,
        headers,
        protocol_version: str = PredictorProtocol.REST_V1.value,
        model_name: str = None,
    ) -> Tuple[Union[Dict, InferRequest], Dict]:
        t1 = time.time()
        attributes = {}
        if isinstance(body, InferRequest):
            return body, attributes
        elif isinstance(body, InferenceRequest) or (
            protocol_version.lower() == PredictorProtocol.REST_V2.value
            and isinstance(body, bytes)
        ):
            return self.decode_inference_request(body, headers, model_name), attributes
        if headers:
            if has_binary_headers(headers):
                # returns CloudEvent
                body = self.get_binary_cloudevent(body, headers)
            elif (
                "content-type" in headers
                and headers["content-type"] not in JSON_HEADERS
            ):
                return body, attributes
        if type(body) is bytes:
            try:
                body = orjson.loads(body)
            except orjson.JSONDecodeError as e:
                raise InvalidInput(f"Unrecognized request format: {e}")

        decoded_body, attributes = self.decode_cloudevent(body)
        t2 = time.time()
        logger.debug(f"decoded request in {round((t2 - t1) * 1000, 9)}ms")
        return decoded_body, attributes

    def decode_cloudevent(self, body) -> Tuple[Union[Dict, InferRequest], Dict]:
        decoded_body = body
        attributes = {}
        if isinstance(body, CloudEvent):
            attributes = body._get_attributes()
            decoded_body = body.get_data()
            try:
                decoded_body = orjson.loads(decoded_body.decode("UTF-8"))
            except (orjson.JSONDecodeError, UnicodeDecodeError) as e:
                # If decoding or parsing failed, check if it was supposed to be JSON UTF-8
                if "content-type" in body._attributes and (
                    body._attributes["content-type"] == "application/cloudevents+json"
                    or body._attributes["content-type"] == "application/json"
                ):
                    raise InvalidInput(
                        f"Failed to decode or parse binary json cloudevent: {e}"
                    )

        elif isinstance(body, dict):
            if is_structured_cloudevent(body):
                decoded_body = body["data"]
                attributes = body
                del attributes["data"]
        return decoded_body, attributes

    def decode_inference_request(
        self, body: Union[bytes, InferenceRequest], headers: Dict, model_name: str
    ) -> InferRequest:
        if isinstance(body, bytes):
            json_length = headers.get(INFERENCE_CONTENT_LENGTH_HEADER, None)
            if json_length is None:
                raise InvalidInput(
                    f"received byte inputs, but the"
                    f"'{INFERENCE_CONTENT_LENGTH_HEADER}' header is missing."
                )
            return InferRequest.from_bytes(body, int(json_length), model_name)
        else:
            return InferRequest.from_inference_request(body, model_name)

    def encode(
        self,
        model_name,
        response,
        headers,
        req_attributes: Dict,
    ) -> Tuple[Dict, Dict[str, str]]:
        response_headers = {}
        # if we received a cloudevent, then also return a cloudevent
        is_cloudevent = False
        is_binary_cloudevent = False
        if isinstance(response, InferResponse):
            response, json_size = response.to_rest()
            if json_size is not None:
                response_headers[INFERENCE_CONTENT_LENGTH_HEADER] = str(json_size)
        if headers:
            if has_binary_headers(headers):
                is_cloudevent = True
                is_binary_cloudevent = True
            if headers.get("content-type", "") == "application/cloudevents+json":
                is_cloudevent = True
        if is_cloudevent:
            response_headers, response = create_response_cloudevent(
                model_name, response, req_attributes, is_binary_cloudevent
            )

            if is_binary_cloudevent:
                response_headers["content-type"] = "application/json"
            else:
                response_headers["content-type"] = "application/cloudevents+json"
        return response, response_headers

    async def infer(
        self,
        model_name: str,
        request: Union[Dict, InferRequest],
        headers: Optional[Dict[str, str]] = None,
    ) -> Tuple[Union[Dict, InferResponse], Dict[str, str]]:
        """Performs inference on the specified model with the provided body and headers.

        If the ``body`` contains an encoded `CloudEvent`_, then it will be decoded and processed.
        The response body/headers will also be encoded as CloudEvents.

        Args:
            model_name (str): Model name.
            request (bytes|Dict): Request body data.
            headers: (Optional[Dict[str, str]]): Request headers.

        Returns:
            Tuple[Union[str, bytes, Dict], Dict[str, str]]:
                - response: The inference result.
                - response_headers: Headers to construct the HTTP response.

        Raises:
            InvalidInput: An error when the body bytes can't be decoded as JSON.

        .. _CloudEvent: https://cloudevents.io/
        """
        # call model locally or remote model workers
        response_headers = {}
        model = await self.get_model(model_name)
        if isinstance(model, OpenAICompletionModel):
            error_msg = f"Model {model_name} is of type OpenAICompletionModel. It does not support the infer method."
            raise InvalidInput(reason=error_msg)
        if not isinstance(model, InferenceModel):
            raise ValueError(
                f"Model of type {type(model).__name__} does not support inference"
            )
        model = cast(InferenceModel, model)
        response, res_headers = await model(request, headers=headers)
        response_headers.update(res_headers)
        return response, response_headers

    async def explain(
        self,
        model_name: str,
        request: Union[bytes, Dict, InferRequest],
        headers: Optional[Dict[str, str]] = None,
    ) -> Tuple[Union[str, bytes, Dict, InferResponse], Dict[str, str]]:
        """Performs explanation for the specified model.

        Args:
            model_name (str): Model name to be used for explanation.
            request (bytes|Dict): Request body data.
            headers: (Optional[Dict[str, str]]): Request headers.

        Returns:
            Dict: Explanation result.

        Raises:
            InvalidInput: An error when the body bytes can't be decoded as JSON.
        """
        # call model locally or remote model workers
        response_headers = headers if headers else {}
        model = await self.get_model(model_name)
        if isinstance(model, OpenAICompletionModel):
            logger.warning(
                f"Model {model_name} is of type OpenAICompletionModel. It does not support the explain method."
                " A request exercised this path and will cause a server crash."
            )
        if not isinstance(model, InferenceModel):
            raise ValueError(
                f"Model of type {type(model).__name__} does not support inference"
            )
        response, res_headers = await model(
            request, verb=InferenceVerb.EXPLAIN, headers=headers
        )
        response_headers.update(res_headers)
        return response, response_headers



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from grpc import ServicerContext

from . import grpc_predict_v2_pb2 as pb
from . import grpc_predict_v2_pb2_grpc
from ..infer_type import InferRequest, InferResponse
from ..dataplane import DataPlane
from ..model_repository_extension import ModelRepositoryExtension
from ...errors import InvalidInput
from ...utils.utils import to_headers


class InferenceServicer(grpc_predict_v2_pb2_grpc.GRPCInferenceServiceServicer):

    def __init__(
        self,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
    ):
        super().__init__()
        self._data_plane = data_plane
        self._mode_repository_extension = model_repository_extension

    @classmethod
    def validate_grpc_request(cls, request: pb.ModelInferRequest):
        raw_inputs_length = len(request.raw_input_contents)
        if raw_inputs_length != 0 and len(request.inputs) != raw_inputs_length:
            raise InvalidInput(
                f"the number of inputs ({len(request.inputs)}) does not match the expected number of "
                f"raw input contents ({raw_inputs_length}) for model '{request.model_name}'."
            )
        if raw_inputs_length != 0:
            for input_ in request.inputs:
                if input_.HasField("contents"):
                    raise InvalidInput(
                        f"contents field must not be specified when using raw_input_contents for input '{input_.name}' for model '{request.model_name}'"
                    )

    async def ServerMetadata(self, request: pb.ServerMetadataRequest, context):
        metadata = self._data_plane.metadata()
        return pb.ServerMetadataResponse(
            name=metadata["name"],
            version=metadata["version"],
            extensions=metadata["extensions"],
        )

    async def ServerLive(
        self, request: pb.ServerLiveRequest, context
    ) -> pb.ServerLiveResponse:
        response = await self._data_plane.live()
        is_live = response["status"] == "alive"
        return pb.ServerLiveResponse(live=is_live)

    async def ServerReady(
        self, request: pb.ServerReadyRequest, context
    ) -> pb.ServerReadyResponse:
        is_ready = await self._data_plane.ready()
        return pb.ServerReadyResponse(ready=is_ready)

    async def ModelReady(
        self, request: pb.ModelReadyRequest, context
    ) -> pb.ModelReadyResponse:
        is_ready = await self._data_plane.model_ready(model_name=request.name)
        return pb.ModelReadyResponse(ready=is_ready)

    async def ModelMetadata(
        self, request: pb.ModelMetadataRequest, context
    ) -> pb.ModelMetadataResponse:
        metadata = await self._data_plane.model_metadata(model_name=request.name)
        return pb.ModelMetadataResponse(
            name=metadata["name"],
            platform=metadata["platform"],
            inputs=metadata["inputs"],
            outputs=metadata["outputs"],
        )

    async def RepositoryModelLoad(
        self, request: pb.RepositoryModelLoadRequest, context
    ) -> pb.RepositoryModelLoadResponse:
        response = await self._mode_repository_extension.load(
            model_name=request.model_name
        )
        return pb.RepositoryModelLoadResponse(
            model_name=response["name"], isLoaded=response["load"]
        )

    async def RepositoryModelUnload(
        self, request: pb.RepositoryModelUnloadRequest, context
    ) -> pb.RepositoryModelUnloadResponse:
        response = await self._mode_repository_extension.unload(
            model_name=request.model_name
        )
        return pb.RepositoryModelUnloadResponse(
            model_name=response["name"], isUnloaded=response["unload"]
        )

    async def ModelInfer(
        self, request: pb.ModelInferRequest, context: ServicerContext
    ) -> pb.ModelInferResponse:
        headers = to_headers(context)
        self.validate_grpc_request(request)
        infer_request = InferRequest.from_grpc(request)
        response_body, _ = await self._data_plane.infer(
            request=infer_request, headers=headers, model_name=request.model_name
        )
        if isinstance(response_body, pb.ModelInferResponse):
            return response_body
        elif isinstance(response_body, InferResponse):
            return response_body.to_grpc()
        else:
            return pb.ModelInferResponse(
                id=response_body["id"],
                model_name=response_body["model_name"],
                outputs=response_body["outputs"],
            )



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging.config
from typing import Optional, Union, Dict

import yaml

from .constants.constants import KSERVE_LOGLEVEL

KSERVE_LOGGER_NAME = "kserve"
KSERVE_TRACE_LOGGER_NAME = "kserve.trace"
KSERVE_LOGGER_FORMAT = (
    "%(asctime)s.%(msecs)03d %(process)s %(name)s "
    "%(levelname)s [%(filename)s:%(funcName)s():%(lineno)s] %(message)s"
)
KSERVE_TRACE_LOGGER_FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(message)s"
KSERVE_LOGGER_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

KSERVE_LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "kserve": {
            "()": "logging.Formatter",
            "fmt": KSERVE_LOGGER_FORMAT,
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
        },
        "kserve_trace": {
            "()": "logging.Formatter",
            "fmt": KSERVE_TRACE_LOGGER_FORMAT,
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
        },
        "uvicorn": {
            "()": "uvicorn.logging.DefaultFormatter",
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
            "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "uvicorn_access": {
            "()": "uvicorn.logging.AccessFormatter",
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
            "fmt": "%(asctime)s.%(msecs)03d %(name)s "
            "%(levelprefix)s %(client_addr)s %(process)s - "
            '"%(request_line)s" %(status_code)s',
            # noqa: E501
        },
    },
    "handlers": {
        "kserve": {
            "formatter": "kserve",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "kserve_trace": {
            "formatter": "kserve_trace",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "uvicorn": {
            "formatter": "uvicorn",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "uvicorn_access": {
            "formatter": "uvicorn_access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "kserve": {
            "handlers": ["kserve"],
            "level": KSERVE_LOGLEVEL,
            "propagate": False,
        },
        "kserve.trace": {
            "handlers": ["kserve_trace"],
            "level": KSERVE_LOGLEVEL,
            "propagate": False,
        },
        "uvicorn": {"handlers": ["uvicorn"], "level": "INFO", "propagate": False},
        "uvicorn.error": {"level": "INFO"},
        "uvicorn.access": {
            "handlers": ["uvicorn_access"],
            "level": "INFO",
            "propagate": False,
        },
    },
}

logger = logging.getLogger(KSERVE_LOGGER_NAME)
trace_logger = logging.getLogger(KSERVE_TRACE_LOGGER_NAME)


def configure_logging(log_config: Optional[Union[Dict, str]] = None):
    """
    Configures Kserve and Uvicorn loggers.
    This function should be called before loading the model / starting the model
    server for consistent logging format.

    :param log_config: (Optional) File path or dict containing log config. If not provided default configuration
                       will be used. If explicitly set to None, the logger will not be configured.
                       - If a dictionary is provided, it will be used directly for configuring the logger.
                       - If a string is provided:
                           - If it ends with '.json', it will be treated as a path to a JSON file containing log
                             configuration.
                           - If it ends with '.yaml' or '.yml', it will be treated as a path to a YAML file containing
                             log configuration.
                           - Otherwise, it will be treated as a path to a configuration file in the format specified in
                             the Python logging module documentation. # See the note about fileConfig() here:
                             # https://docs.python.org/3/library/logging.config.html#configuration-file-format
    """
    if log_config is None:
        logging.config.dictConfig(KSERVE_LOG_CONFIG)
    elif isinstance(log_config, dict):
        logging.config.dictConfig(log_config)
    elif log_config.endswith(".json"):
        with open(log_config) as file:
            loaded_config = json.load(file)
            logging.config.dictConfig(loaded_config)
    elif log_config.endswith((".yaml", ".yml")):
        with open(log_config) as file:
            loaded_config = yaml.safe_load(file)
            logging.config.dictConfig(loaded_config)
    else:
        # See the note about fileConfig() here:
        # https://docs.python.org/3/library/logging.config.html#configuration-file-format
        logging.config.fileConfig(log_config, disable_existing_loggers=False)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, List, Union, Dict

import orjson
import pydantic

from pydantic import BaseModel, StrictBool, StrictInt, StrictFloat

# this is needed to add support for both pydantic 1.x and 2.x
is_pydantic_2 = pydantic.__version__.startswith("2.")

if is_pydantic_2:
    from pydantic import ConfigDict

# TODO: in the future, this file can be auto generated
# https://pydantic-docs.helpmanual.io/datamodel_code_generator/

# Reference: https://github.com/kserve/kserve/blob/master/docs/predict-api/v2/rest_predict_v2.yaml


InferParameter = Union[StrictFloat, StrictInt, StrictBool, str]
Parameters = Dict[str, InferParameter]


server_live_response_schema_extra = {"example": {"live": True}}


class ServerLiveResponse(BaseModel):
    live: bool

    if is_pydantic_2:
        model_config = ConfigDict(json_schema_extra=server_live_response_schema_extra)

    else:

        class Config:
            schema_extra = server_live_response_schema_extra


server_ready_response_schema_extra = {"example": {"ready": True}}


class ServerReadyResponse(BaseModel):
    ready: bool

    if is_pydantic_2:
        model_config = ConfigDict(json_schema_extra=server_ready_response_schema_extra)

    else:

        class Config:
            schema_extra = server_ready_response_schema_extra


class ServerMetadataResponse(BaseModel):
    """ServerMetadataResponse

    $server_metadata_response =
    {
      "name" : $string,
      "version" : $string,
      "extensions" : [ $string, ... ]
    }
    """

    name: str
    version: str
    extensions: List[str]


class MetadataTensor(BaseModel):
    """MetadataTensor

    $metadata_tensor =
    {
      "name" : $string,
      "datatype" : $string,
      "shape" : [ $number, ... ]
    }
    """

    name: str
    datatype: str
    shape: List[int]


class ListModelsResponse(BaseModel):
    """ListModelsResponse

    $models_list_response =
    {
      "models" : [ $string, ... ]
    }
    """

    models: List[str]


class ModelMetadataResponse(BaseModel):
    """ModelMetadataResponse

    $metadata_model_response =
    {
      "name" : $string,
      "versions" : [ $string, ... ] #optional,
      "platform" : $string,
      "inputs" : [ $metadata_tensor, ... ],
      "outputs" : [ $metadata_tensor, ... ]
    }
    """

    name: str
    versions: Optional[List[str]] = None
    platform: str
    inputs: List[MetadataTensor]
    outputs: List[MetadataTensor]


class ModelReadyResponse(BaseModel):
    """ModelReadyResponse

    $ready_model_response =
    {
      "name": $string,
      "ready": $bool
    }
    """

    name: str
    ready: bool


class RequestInput(BaseModel):
    """RequestInput Model

    $request_input =
    {
      "name" : $string,
      "shape" : [ $number, ... ],
      "datatype"  : $string,
      "parameters" : $parameters #optional,
      "data" : $tensor_data
    }
    """

    name: str
    shape: List[int]
    datatype: str
    parameters: Optional[Parameters] = None
    data: List


class RequestOutput(BaseModel):
    """RequestOutput Model

    $request_output =
    {
      "name" : $string,
      "parameters" : $parameters #optional,
    }
    """

    name: str
    parameters: Optional[Parameters] = None


class ResponseOutput(BaseModel):
    """ResponseOutput Model

    $response_output =
    {
      "name" : $string,
      "shape" : [ $number, ... ],
      "datatype"  : $string,
      "parameters" : $parameters #optional,
      "data" : $tensor_data
    }
    """

    name: str
    shape: List[int]
    datatype: str
    parameters: Optional[Parameters] = None
    data: List


inference_request_schema_extra = {
    "example": {
        "id": "42",
        "inputs": [
            {
                "name": "input0",
                "shape": [2, 2],
                "datatype": "UINT32",
                "data": [1, 2, 3, 4],
            },
            {"name": "input1", "shape": [3], "datatype": "BOOL", "data": ["true"]},
        ],
        "outputs": [{"name": "output0"}],
    }
}


class InferenceRequest(BaseModel):
    """InferenceRequest Model

    $inference_request =
    {
      "id" : $string #optional,
      "parameters" : $parameters #optional,
      "inputs" : [ $request_input, ... ],
      "outputs" : [ $request_output, ... ] #optional
    }
    """

    id: Optional[str] = None
    parameters: Optional[Parameters] = None
    inputs: List[RequestInput]
    outputs: Optional[List[RequestOutput]] = None

    if is_pydantic_2:
        model_config = ConfigDict(json_schema_extra=inference_request_schema_extra)

    else:

        class Config:
            json_loads = orjson.loads
            schema_extra = inference_request_schema_extra


inference_response_schema_extra = {
    "example": {
        "id": "42",
        "outputs": [
            {
                "name": "output0",
                "shape": [3, 2],
                "datatype": "FP32",
                "data": [1.0, 1.1, 2.0, 2.1, 3.0, 3.1],
            }
        ],
    }
}


class InferenceResponse(BaseModel):
    """InferenceResponse

    $inference_response =
    {
      "model_name" : $string,
      "model_version" : $string #optional,
      "id" : $string,
      "parameters" : $parameters #optional,
      "outputs" : [ $response_output, ... ]
    }
    """

    model_name: str
    model_version: Optional[str] = None
    id: str
    parameters: Optional[Parameters] = None
    outputs: List[ResponseOutput]

    if is_pydantic_2:
        model_config = ConfigDict(
            protected_namespaces=(), json_schema_extra=inference_response_schema_extra
        )
    else:

        class Config:
            schema_extra = inference_response_schema_extra



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from fastapi import FastAPI

from ....model import Model
from ....model_repository import ModelRepository


def get_open_ai_models(repository: ModelRepository) -> List[Model]:
    """Retrieve all models in the repository that implement the OpenAI interface"""
    from .openai_model import OpenAIModel

    return [
        model
        for _, model in repository.get_models().items()
        if isinstance(model, OpenAIModel)
    ]


def maybe_register_openai_endpoints(app: FastAPI, model_registry: ModelRepository):
    open_ai_models = get_open_ai_models(model_registry)
    # If no OpenAI models then no need to add the endpoints
    if len(open_ai_models) == 0:
        return
    from .dataplane import OpenAIDataPlane
    from .endpoints import register_openai_endpoints

    # Create a model repository with just the OpenAI models
    openai_model_registry = ModelRepository()
    for model in open_ai_models:
        openai_model_registry.update(model)

    # Add the OpenAI completion and chat completion endpoints.
    register_openai_endpoints(app, OpenAIDataPlane(openai_model_registry))



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from .model import Model
from .model_server import ModelServer
from .inference_client import InferenceGRPCClient, InferenceRESTClient, RESTConfig
from .protocol.infer_type import InferRequest, InferInput, InferResponse, InferOutput
from .model_repository import ModelRepository
from .constants import constants
from .utils import utils

# import client apis into kserve package
from .api.kserve_client import KServeClient
from .constants import constants

# import ApiClient
from .api_client import ApiClient
from .configuration import Configuration
from .exceptions import OpenApiException
from .exceptions import ApiTypeError
from .exceptions import ApiValueError
from .exceptions import ApiKeyError
from .exceptions import ApiException

# import v1alpha1 models into kserve packages
from .models.v1alpha1_built_in_adapter import V1alpha1BuiltInAdapter
from .models.v1alpha1_cluster_serving_runtime import V1alpha1ClusterServingRuntime
from .models.v1alpha1_cluster_serving_runtime_list import (
    V1alpha1ClusterServingRuntimeList,
)
from .models.v1alpha1_container import V1alpha1Container
from .models.v1alpha1_inference_graph import V1alpha1InferenceGraph
from .models.v1alpha1_inference_graph_list import V1alpha1InferenceGraphList
from .models.v1alpha1_inference_graph_spec import V1alpha1InferenceGraphSpec
from .models.v1alpha1_inference_graph_status import V1alpha1InferenceGraphStatus
from .models.v1alpha1_inference_router import V1alpha1InferenceRouter
from .models.v1alpha1_inference_step import V1alpha1InferenceStep
from .models.v1alpha1_inference_target import V1alpha1InferenceTarget
from .models.v1alpha1_model_spec import V1alpha1ModelSpec
from .models.v1alpha1_serving_runtime import V1alpha1ServingRuntime
from .models.v1alpha1_serving_runtime_list import V1alpha1ServingRuntimeList
from .models.v1alpha1_serving_runtime_pod_spec import V1alpha1ServingRuntimePodSpec
from .models.v1alpha1_serving_runtime_spec import V1alpha1ServingRuntimeSpec
from .models.v1alpha1_storage_helper import V1alpha1StorageHelper
from .models.v1alpha1_supported_model_format import V1alpha1SupportedModelFormat
from .models.v1alpha1_trained_model import V1alpha1TrainedModel
from .models.v1alpha1_trained_model_list import V1alpha1TrainedModelList
from .models.v1alpha1_trained_model_spec import V1alpha1TrainedModelSpec

# import v1beta1 models into sdk package
from .models.knative_addressable import KnativeAddressable
from .models.knative_condition import KnativeCondition
from .models.knative_url import KnativeURL
from .models.knative_volatile_time import KnativeVolatileTime
from .models.net_url_userinfo import NetUrlUserinfo
from .models.v1beta1_art_explainer_spec import V1beta1ARTExplainerSpec
from .models.v1beta1_batcher import V1beta1Batcher
from .models.v1beta1_component_extension_spec import V1beta1ComponentExtensionSpec
from .models.v1beta1_component_status_spec import V1beta1ComponentStatusSpec
from .models.v1beta1_custom_explainer import V1beta1CustomExplainer
from .models.v1beta1_custom_predictor import V1beta1CustomPredictor
from .models.v1beta1_custom_transformer import V1beta1CustomTransformer
from .models.v1beta1_deploy_config import V1beta1DeployConfig
from .models.v1beta1_explainer_config import V1beta1ExplainerConfig
from .models.v1beta1_explainer_extension_spec import V1beta1ExplainerExtensionSpec
from .models.v1beta1_explainer_spec import V1beta1ExplainerSpec
from .models.v1beta1_explainers_config import V1beta1ExplainersConfig
from .models.v1beta1_inference_service import V1beta1InferenceService
from .models.v1beta1_inference_service_list import V1beta1InferenceServiceList
from .models.v1beta1_inference_service_spec import V1beta1InferenceServiceSpec
from .models.v1beta1_inference_service_status import V1beta1InferenceServiceStatus
from .models.v1beta1_inference_services_config import V1beta1InferenceServicesConfig
from .models.v1beta1_ingress_config import V1beta1IngressConfig
from .models.v1beta1_light_gbm_spec import V1beta1LightGBMSpec
from .models.v1beta1_logger_spec import V1beta1LoggerSpec
from .models.v1beta1_model_format import V1beta1ModelFormat
from .models.v1beta1_model_spec import V1beta1ModelSpec
from .models.v1beta1_onnx_runtime_spec import V1beta1ONNXRuntimeSpec
from .models.v1beta1_pmml_spec import V1beta1PMMLSpec
from .models.v1beta1_paddle_server_spec import V1beta1PaddleServerSpec
from .models.v1beta1_pod_spec import V1beta1PodSpec
from .models.v1beta1_predictor_config import V1beta1PredictorConfig
from .models.v1beta1_predictor_extension_spec import V1beta1PredictorExtensionSpec
from .models.v1beta1_predictor_protocols import V1beta1PredictorProtocols
from .models.v1beta1_predictor_spec import V1beta1PredictorSpec
from .models.v1beta1_predictors_config import V1beta1PredictorsConfig
from .models.v1beta1_sk_learn_spec import V1beta1SKLearnSpec
from .models.v1beta1_tf_serving_spec import V1beta1TFServingSpec
from .models.v1beta1_torch_serve_spec import V1beta1TorchServeSpec
from .models.v1beta1_transformer_config import V1beta1TransformerConfig
from .models.v1beta1_transformer_spec import V1beta1TransformerSpec
from .models.v1beta1_transformers_config import V1beta1TransformersConfig
from .models.v1beta1_triton_spec import V1beta1TritonSpec
from .models.v1beta1_xg_boost_spec import V1beta1XGBoostSpec
from .models.v1beta1_storage_spec import V1beta1StorageSpec



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict, Union

from fastapi import FastAPI, APIRouter
from fastapi.requests import Request
from fastapi.responses import Response

from .v2_datamodels import (
    InferenceRequest,
    ServerMetadataResponse,
    ServerLiveResponse,
    ServerReadyResponse,
    ModelMetadataResponse,
    InferenceResponse,
    ModelReadyResponse,
    ListModelsResponse,
)
from ..dataplane import DataPlane
from ..model_repository_extension import ModelRepositoryExtension
from ...constants.constants import V2_ROUTE_PREFIX, PredictorProtocol
from ...errors import ModelNotReady, ServerNotLive, ServerNotReady


class V2Endpoints:
    """KServe V2 Endpoints"""

    def __init__(
        self,
        dataplane: DataPlane,
        model_repository_extension: Optional[ModelRepositoryExtension] = None,
    ):
        self.model_repository_extension = model_repository_extension
        self.dataplane = dataplane

    async def metadata(self) -> ServerMetadataResponse:
        """Server metadata endpoint.

        Returns:
            ServerMetadataResponse: Server metadata JSON object.
        """
        return ServerMetadataResponse.parse_obj(self.dataplane.metadata())

    async def live(self) -> ServerLiveResponse:
        """Server live endpoint.

        Returns:
            ServerLiveResponse: Server live message.
        """
        response = await self.dataplane.live()
        is_live = response["status"] == "alive"
        if not is_live:
            raise ServerNotLive()
        return ServerLiveResponse(live=is_live)

    async def ready(self) -> ServerReadyResponse:
        """Server ready endpoint.

        Returns:
            ServerReadyResponse: Server ready message.
        """
        is_ready = await self.dataplane.ready()
        if not is_ready:
            raise ServerNotReady()
        return ServerReadyResponse(ready=is_ready)

    async def models(self) -> ListModelsResponse:
        """Get a list of models in the model registry.

        Returns:
            ListModelsResponse: List of models object.
        """
        models = list(self.dataplane.model_registry.get_models().keys())
        return ListModelsResponse.parse_obj({"models": models})

    async def model_metadata(
        self, model_name: str, model_version: Optional[str] = None
    ) -> ModelMetadataResponse:
        """Model metadata handler. It provides information about a model.

        Args:
            model_name (str): Model name.
            model_version (Optional[str]): Model version (optional).

        Returns:
            ModelMetadataResponse: Model metadata object.
        """
        # TODO: support model_version
        if model_version:
            raise NotImplementedError("Model versioning not supported yet.")

        metadata = await self.dataplane.model_metadata(model_name)
        return ModelMetadataResponse.parse_obj(metadata)

    async def model_ready(
        self, model_name: str, model_version: Optional[str] = None
    ) -> ModelReadyResponse:
        """Check if a given model is ready.

        Args:
            model_name (str): Model name.
            model_version (str): Model version.

        Returns:
            ModelReadyResponse: Model ready object
        """
        # TODO: support model_version
        if model_version:
            raise NotImplementedError("Model versioning not supported yet.")

        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        return ModelReadyResponse.parse_obj({"name": model_name, "ready": model_ready})

    async def infer(
        self,
        raw_request: Request,
        raw_response: Response,
        model_name: str,
        request_body: Union[InferenceRequest, bytes],
        model_version: Optional[str] = None,
    ) -> Union[InferenceResponse, Response]:
        """Infer handler.

        Args:
            raw_request (Request): fastapi request object,
            raw_response (Response): fastapi response object,
            model_name (str): Model name.
            request_body (InferenceRequest): Inference request body.
            model_version (Optional[str]): Model version (optional).

        Returns:
            InferenceResponse: Inference response object.
        """
        # TODO: support model_version
        if model_version:
            raise NotImplementedError("Model versioning not supported yet.")

        model_ready = await self.dataplane.model_ready(model_name)

        if not model_ready:
            raise ModelNotReady(model_name)

        request_headers = dict(raw_request.headers)

        infer_request, _ = self.dataplane.decode(
            request_body,
            request_headers,
            protocol_version=PredictorProtocol.REST_V2.value,
            model_name=model_name,
        )
        response, response_headers = await self.dataplane.infer(
            model_name=model_name,
            request=infer_request,
            headers=request_headers,
        )
        response, res_headers = self.dataplane.encode(
            model_name=model_name,
            response=response,
            headers=request_headers,
            req_attributes={},
        )

        response_headers.update(res_headers)
        response_headers.pop("content-length", None)
        response_headers.pop("content-type", None)

        if response_headers:
            raw_response.headers.update(response_headers)
        if isinstance(response, bytes):
            raw_response.status_code = 200
            raw_response.body = response
            res = raw_response
        else:
            res = InferenceResponse.parse_obj(response)
        return res

    async def load(self, model_name: str) -> Dict:
        """Model load handler.

        Args:
            model_name (str): Model name.

        Returns:
            Dict: {"name": model_name, "load": True}
        """
        await self.model_repository_extension.load(model_name)
        return {"name": model_name, "load": True}

    async def unload(self, model_name: str) -> Dict:
        """Model unload handler.

        Args:
            model_name (str): Model name.

        Returns:
            Dict: {"name": model_name, "unload": True}
        """
        await self.model_repository_extension.unload(model_name)
        return {"name": model_name, "unload": True}


def register_v2_endpoints(
    app: FastAPI,
    dataplane: DataPlane,
    model_repository_extension: Optional[ModelRepositoryExtension],
):
    """Register V2 endpoints.

    Args:
        app (FastAPI): FastAPI app.
        dataplane (DataPlane): DataPlane object.
        model_repository_extension (Optional[ModelRepositoryExtension]): Model repository extension.
    """
    v2_endpoints = V2Endpoints(
        dataplane=dataplane, model_repository_extension=model_repository_extension
    )
    v2_router = APIRouter(prefix=V2_ROUTE_PREFIX, tags=["V2"])
    v2_router.add_api_route(
        r"",
        v2_endpoints.metadata,
        response_model=ServerMetadataResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/health/live",
        v2_endpoints.live,
        response_model=ServerLiveResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/health/ready",
        v2_endpoints.ready,
        response_model=ServerReadyResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/models",
        v2_endpoints.models,
        response_model=ListModelsResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/models/{model_name}",
        v2_endpoints.model_metadata,
        response_model=ModelMetadataResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/models/{model_name}/versions/{model_version}",
        v2_endpoints.model_metadata,
        response_model=ModelMetadataResponse,
        methods=["GET"],
        include_in_schema=False,
    )
    v2_router.add_api_route(
        r"/models/{model_name}/ready",
        v2_endpoints.model_ready,
        response_model=ModelReadyResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/models/{model_name}/versions/{model_version}/ready",
        v2_endpoints.model_ready,
        response_model=ModelReadyResponse,
        methods=["GET"],
    )
    v2_router.add_api_route(
        r"/models/{model_name}/infer",
        v2_endpoints.infer,
        response_model=None,
        methods=["POST"],
    )
    v2_router.add_api_route(
        r"/models/{model_name}/versions/{model_version}/infer",
        v2_endpoints.infer,
        response_model=None,
        methods=["POST"],
        include_in_schema=False,
    )
    v2_router.add_api_route(
        r"/repository/models/{model_name}/load", v2_endpoints.load, methods=["POST"]
    )
    v2_router.add_api_route(
        r"/repository/models/{model_name}/unload", v2_endpoints.unload, methods=["POST"]
    )
    app.include_router(v2_router)



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1beta1IngressConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'additional_ingress_domains': 'list[str]',
        'disable_ingress_creation': 'bool',
        'disable_istio_virtual_host': 'bool',
        'domain_template': 'str',
        'enable_gateway_api': 'bool',
        'ingress_class_name': 'str',
        'ingress_domain': 'str',
        'ingress_gateway': 'str',
        'knative_local_gateway_service': 'str',
        'kserve_ingress_gateway': 'str',
        'local_gateway': 'str',
        'local_gateway_service': 'str',
        'path_template': 'str',
        'url_scheme': 'str'
    }

    attribute_map = {
        'additional_ingress_domains': 'additionalIngressDomains',
        'disable_ingress_creation': 'disableIngressCreation',
        'disable_istio_virtual_host': 'disableIstioVirtualHost',
        'domain_template': 'domainTemplate',
        'enable_gateway_api': 'enableGatewayApi',
        'ingress_class_name': 'ingressClassName',
        'ingress_domain': 'ingressDomain',
        'ingress_gateway': 'ingressGateway',
        'knative_local_gateway_service': 'knativeLocalGatewayService',
        'kserve_ingress_gateway': 'kserveIngressGateway',
        'local_gateway': 'localGateway',
        'local_gateway_service': 'localGatewayService',
        'path_template': 'pathTemplate',
        'url_scheme': 'urlScheme'
    }

    def __init__(self, additional_ingress_domains=None, disable_ingress_creation=None, disable_istio_virtual_host=None, domain_template=None, enable_gateway_api=None, ingress_class_name=None, ingress_domain=None, ingress_gateway=None, knative_local_gateway_service=None, kserve_ingress_gateway=None, local_gateway=None, local_gateway_service=None, path_template=None, url_scheme=None, local_vars_configuration=None):  # noqa: E501
        """V1beta1IngressConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._additional_ingress_domains = None
        self._disable_ingress_creation = None
        self._disable_istio_virtual_host = None
        self._domain_template = None
        self._enable_gateway_api = None
        self._ingress_class_name = None
        self._ingress_domain = None
        self._ingress_gateway = None
        self._knative_local_gateway_service = None
        self._kserve_ingress_gateway = None
        self._local_gateway = None
        self._local_gateway_service = None
        self._path_template = None
        self._url_scheme = None
        self.discriminator = None

        if additional_ingress_domains is not None:
            self.additional_ingress_domains = additional_ingress_domains
        if disable_ingress_creation is not None:
            self.disable_ingress_creation = disable_ingress_creation
        if disable_istio_virtual_host is not None:
            self.disable_istio_virtual_host = disable_istio_virtual_host
        if domain_template is not None:
            self.domain_template = domain_template
        if enable_gateway_api is not None:
            self.enable_gateway_api = enable_gateway_api
        if ingress_class_name is not None:
            self.ingress_class_name = ingress_class_name
        if ingress_domain is not None:
            self.ingress_domain = ingress_domain
        if ingress_gateway is not None:
            self.ingress_gateway = ingress_gateway
        if knative_local_gateway_service is not None:
            self.knative_local_gateway_service = knative_local_gateway_service
        if kserve_ingress_gateway is not None:
            self.kserve_ingress_gateway = kserve_ingress_gateway
        if local_gateway is not None:
            self.local_gateway = local_gateway
        if local_gateway_service is not None:
            self.local_gateway_service = local_gateway_service
        if path_template is not None:
            self.path_template = path_template
        if url_scheme is not None:
            self.url_scheme = url_scheme

    @property
    def additional_ingress_domains(self):
        """Gets the additional_ingress_domains of this V1beta1IngressConfig.  # noqa: E501


        :return: The additional_ingress_domains of this V1beta1IngressConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._additional_ingress_domains

    @additional_ingress_domains.setter
    def additional_ingress_domains(self, additional_ingress_domains):
        """Sets the additional_ingress_domains of this V1beta1IngressConfig.


        :param additional_ingress_domains: The additional_ingress_domains of this V1beta1IngressConfig.  # noqa: E501
        :type: list[str]
        """

        self._additional_ingress_domains = additional_ingress_domains

    @property
    def disable_ingress_creation(self):
        """Gets the disable_ingress_creation of this V1beta1IngressConfig.  # noqa: E501


        :return: The disable_ingress_creation of this V1beta1IngressConfig.  # noqa: E501
        :rtype: bool
        """
        return self._disable_ingress_creation

    @disable_ingress_creation.setter
    def disable_ingress_creation(self, disable_ingress_creation):
        """Sets the disable_ingress_creation of this V1beta1IngressConfig.


        :param disable_ingress_creation: The disable_ingress_creation of this V1beta1IngressConfig.  # noqa: E501
        :type: bool
        """

        self._disable_ingress_creation = disable_ingress_creation

    @property
    def disable_istio_virtual_host(self):
        """Gets the disable_istio_virtual_host of this V1beta1IngressConfig.  # noqa: E501


        :return: The disable_istio_virtual_host of this V1beta1IngressConfig.  # noqa: E501
        :rtype: bool
        """
        return self._disable_istio_virtual_host

    @disable_istio_virtual_host.setter
    def disable_istio_virtual_host(self, disable_istio_virtual_host):
        """Sets the disable_istio_virtual_host of this V1beta1IngressConfig.


        :param disable_istio_virtual_host: The disable_istio_virtual_host of this V1beta1IngressConfig.  # noqa: E501
        :type: bool
        """

        self._disable_istio_virtual_host = disable_istio_virtual_host

    @property
    def domain_template(self):
        """Gets the domain_template of this V1beta1IngressConfig.  # noqa: E501


        :return: The domain_template of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._domain_template

    @domain_template.setter
    def domain_template(self, domain_template):
        """Sets the domain_template of this V1beta1IngressConfig.


        :param domain_template: The domain_template of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._domain_template = domain_template

    @property
    def enable_gateway_api(self):
        """Gets the enable_gateway_api of this V1beta1IngressConfig.  # noqa: E501


        :return: The enable_gateway_api of this V1beta1IngressConfig.  # noqa: E501
        :rtype: bool
        """
        return self._enable_gateway_api

    @enable_gateway_api.setter
    def enable_gateway_api(self, enable_gateway_api):
        """Sets the enable_gateway_api of this V1beta1IngressConfig.


        :param enable_gateway_api: The enable_gateway_api of this V1beta1IngressConfig.  # noqa: E501
        :type: bool
        """

        self._enable_gateway_api = enable_gateway_api

    @property
    def ingress_class_name(self):
        """Gets the ingress_class_name of this V1beta1IngressConfig.  # noqa: E501


        :return: The ingress_class_name of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._ingress_class_name

    @ingress_class_name.setter
    def ingress_class_name(self, ingress_class_name):
        """Sets the ingress_class_name of this V1beta1IngressConfig.


        :param ingress_class_name: The ingress_class_name of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._ingress_class_name = ingress_class_name

    @property
    def ingress_domain(self):
        """Gets the ingress_domain of this V1beta1IngressConfig.  # noqa: E501


        :return: The ingress_domain of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._ingress_domain

    @ingress_domain.setter
    def ingress_domain(self, ingress_domain):
        """Sets the ingress_domain of this V1beta1IngressConfig.


        :param ingress_domain: The ingress_domain of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._ingress_domain = ingress_domain

    @property
    def ingress_gateway(self):
        """Gets the ingress_gateway of this V1beta1IngressConfig.  # noqa: E501


        :return: The ingress_gateway of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._ingress_gateway

    @ingress_gateway.setter
    def ingress_gateway(self, ingress_gateway):
        """Sets the ingress_gateway of this V1beta1IngressConfig.


        :param ingress_gateway: The ingress_gateway of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._ingress_gateway = ingress_gateway

    @property
    def knative_local_gateway_service(self):
        """Gets the knative_local_gateway_service of this V1beta1IngressConfig.  # noqa: E501


        :return: The knative_local_gateway_service of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._knative_local_gateway_service

    @knative_local_gateway_service.setter
    def knative_local_gateway_service(self, knative_local_gateway_service):
        """Sets the knative_local_gateway_service of this V1beta1IngressConfig.


        :param knative_local_gateway_service: The knative_local_gateway_service of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._knative_local_gateway_service = knative_local_gateway_service

    @property
    def kserve_ingress_gateway(self):
        """Gets the kserve_ingress_gateway of this V1beta1IngressConfig.  # noqa: E501


        :return: The kserve_ingress_gateway of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._kserve_ingress_gateway

    @kserve_ingress_gateway.setter
    def kserve_ingress_gateway(self, kserve_ingress_gateway):
        """Sets the kserve_ingress_gateway of this V1beta1IngressConfig.


        :param kserve_ingress_gateway: The kserve_ingress_gateway of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._kserve_ingress_gateway = kserve_ingress_gateway

    @property
    def local_gateway(self):
        """Gets the local_gateway of this V1beta1IngressConfig.  # noqa: E501


        :return: The local_gateway of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._local_gateway

    @local_gateway.setter
    def local_gateway(self, local_gateway):
        """Sets the local_gateway of this V1beta1IngressConfig.


        :param local_gateway: The local_gateway of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._local_gateway = local_gateway

    @property
    def local_gateway_service(self):
        """Gets the local_gateway_service of this V1beta1IngressConfig.  # noqa: E501


        :return: The local_gateway_service of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._local_gateway_service

    @local_gateway_service.setter
    def local_gateway_service(self, local_gateway_service):
        """Sets the local_gateway_service of this V1beta1IngressConfig.


        :param local_gateway_service: The local_gateway_service of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._local_gateway_service = local_gateway_service

    @property
    def path_template(self):
        """Gets the path_template of this V1beta1IngressConfig.  # noqa: E501


        :return: The path_template of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._path_template

    @path_template.setter
    def path_template(self, path_template):
        """Sets the path_template of this V1beta1IngressConfig.


        :param path_template: The path_template of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._path_template = path_template

    @property
    def url_scheme(self):
        """Gets the url_scheme of this V1beta1IngressConfig.  # noqa: E501


        :return: The url_scheme of this V1beta1IngressConfig.  # noqa: E501
        :rtype: str
        """
        return self._url_scheme

    @url_scheme.setter
    def url_scheme(self, url_scheme):
        """Sets the url_scheme of this V1beta1IngressConfig.


        :param url_scheme: The url_scheme of this V1beta1IngressConfig.  # noqa: E501
        :type: str
        """

        self._url_scheme = url_scheme

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1beta1IngressConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1beta1IngressConfig):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .openai_model import (
    ChatPrompt,
    CompletionRequest,
    ChatCompletionRequest,
    EmbeddingRequest,
    OpenAIModel,
    OpenAICompletionModel,
    OpenAIEmbeddingModel,
)

from .openai_proxy_model import OpenAIProxyModel
from .openai_chat_adapter_model import OpenAIChatAdapterModel
from .types import ChatCompletionRequestMessage

__all__ = [
    "OpenAIModel",
    "OpenAIChatAdapterModel",
    "OpenAICompletionModel",
    "OpenAIEmbeddingModel",
    "OpenAIProxyModel",
    "ChatPrompt",
    "CompletionRequest",
    "ChatCompletionRequest",
    "ChatCompletionRequestMessage",
    "EmbeddingRequest",
]
